"""add generic taxonomy models

Revision ID: 3baf42d251a6
Revises: 2f3c1a2d6b10
Create Date: 2025-08-05 21:49:39.619569

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "3baf42d251a6"
down_revision = "2f3c1a2d6b10"
branch_labels = None
depends_on = None


def upgrade():
    # Create taxonomy table
    op.create_table(
        "taxonomy",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("fides_key", sa.String(), nullable=False),
        sa.Column("organization_fides_key", sa.Text(), nullable=True),
        sa.Column("tags", postgresql.ARRAY(sa.String()), nullable=True),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("fides_key"),
        sa.UniqueConstraint("id", name="uq_taxonomy_id"),
    )
    op.create_index(
        op.f("ix_taxonomy_fides_key"), "taxonomy", ["fides_key"], unique=True
    )

    # Create taxonomy_allowed_usage table
    op.create_table(
        "taxonomy_allowed_usage",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("source_taxonomy_key", sa.String(), nullable=False),
        sa.Column(
            "target_type", sa.String(), nullable=False
        ),  # Can be "system" OR "data_categories"
        sa.ForeignKeyConstraint(
            ["source_taxonomy_key"], ["taxonomy.fides_key"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("source_taxonomy_key", "target_type"),
        sa.UniqueConstraint("id", name="uq_taxonomy_allowed_usage_id"),
    )

    # Create taxonomy_element table
    op.create_table(
        "taxonomy_element",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("fides_key", sa.String(), nullable=False),
        sa.Column("organization_fides_key", sa.Text(), nullable=True),
        sa.Column("tags", postgresql.ARRAY(sa.String()), nullable=True),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("taxonomy_type", sa.String(), nullable=False),
        sa.Column("parent_key", sa.Text(), nullable=True),
        sa.Column("active", sa.BOOLEAN(), nullable=False),
        sa.ForeignKeyConstraint(
            ["parent_key"], ["taxonomy_element.fides_key"], ondelete="RESTRICT"
        ),
        sa.ForeignKeyConstraint(
            ["taxonomy_type"], ["taxonomy.fides_key"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("fides_key"),
        sa.UniqueConstraint("id", name="uq_taxonomy_element_id"),
    )
    op.create_index(
        op.f("ix_taxonomy_element_fides_key"),
        "taxonomy_element",
        ["fides_key"],
        unique=True,
    )
    op.create_index(
        op.f("ix_taxonomy_element_active"), "taxonomy_element", ["active"], unique=False
    )
    op.create_index(
        op.f("ix_taxonomy_element_parent_key"),
        "taxonomy_element",
        ["parent_key"],
        unique=False,
    )
    op.create_index(
        op.f("ix_taxonomy_element_taxonomy_type"),
        "taxonomy_element",
        ["taxonomy_type"],
        unique=False,
    )

    # Create taxonomy_usage table
    op.create_table(
        "taxonomy_usage",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("source_element_key", sa.String(), nullable=False),
        sa.Column("target_element_key", sa.String(), nullable=False),
        sa.Column("source_taxonomy", sa.String(), nullable=False),
        sa.Column("target_taxonomy", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["source_taxonomy", "target_taxonomy"],
            [
                "taxonomy_allowed_usage.source_taxonomy_key",
                "taxonomy_allowed_usage.target_type",
            ],
            name="fk_taxonomy_usage_allowed",
            ondelete="RESTRICT",
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "source_element_key",
            "target_element_key",
            name="uq_taxonomy_usage",
        ),
    )
    op.create_index(
        op.f("ix_taxonomy_usage_source_element_key"),
        "taxonomy_usage",
        ["source_element_key"],
        unique=False,
    )
    op.create_index(
        op.f("ix_taxonomy_usage_id"), "taxonomy_usage", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_taxonomy_usage_source_taxonomy"),
        "taxonomy_usage",
        ["source_taxonomy"],
        unique=False,
    )
    op.create_index(
        op.f("ix_taxonomy_usage_target_element_key"),
        "taxonomy_usage",
        ["target_element_key"],
        unique=False,
    )
    op.create_index(
        op.f("ix_taxonomy_usage_target_taxonomy"),
        "taxonomy_usage",
        ["target_taxonomy"],
        unique=False,
    )

    # Seed legacy taxonomies that are managed by the system
    op.execute(
        """
        INSERT INTO taxonomy (id, created_at, updated_at, fides_key, organization_fides_key, tags, name, description)
        VALUES
          ('data_category', now(), now(), 'data_category', NULL, NULL, 'Data categories', 'Taxonomy for data categories'),
          ('data_use', now(), now(), 'data_use', NULL, NULL, 'Data uses', 'Taxonomy for data uses'),
          ('data_subject', now(), now(), 'data_subject', NULL, NULL, 'Data subjects', 'Taxonomy for data subjects'),
          ('system_group', now(), now(), 'system_group', NULL, NULL, 'System groups', 'Taxonomy for system groups')
        """
    )


def downgrade():
    # Drop taxonomy_usage
    op.drop_index(
        op.f("ix_taxonomy_usage_source_element_key"), table_name="taxonomy_usage"
    )
    op.drop_index(
        op.f("ix_taxonomy_usage_target_taxonomy"), table_name="taxonomy_usage"
    )
    op.drop_index(
        op.f("ix_taxonomy_usage_target_element_key"), table_name="taxonomy_usage"
    )
    op.drop_index(
        op.f("ix_taxonomy_usage_source_taxonomy"), table_name="taxonomy_usage"
    )
    op.drop_index(op.f("ix_taxonomy_usage_id"), table_name="taxonomy_usage")
    op.drop_table("taxonomy_usage")

    # Drop taxonomy_element
    op.drop_index(
        op.f("ix_taxonomy_element_taxonomy_type"), table_name="taxonomy_element"
    )
    op.drop_index(op.f("ix_taxonomy_element_parent_key"), table_name="taxonomy_element")
    op.drop_index(op.f("ix_taxonomy_element_active"), table_name="taxonomy_element")
    op.drop_index(op.f("ix_taxonomy_element_fides_key"), table_name="taxonomy_element")
    op.drop_table("taxonomy_element")

    # Drop taxonomy_allowed_usage
    op.drop_table("taxonomy_allowed_usage")

    # Drop taxonomy
    op.drop_index(op.f("ix_taxonomy_fides_key"), table_name="taxonomy")
    op.drop_table("taxonomy")



from enum import Enum


class IntegrationFeature(str, Enum):
    """
    Features that can be enabled for different integration types.
    These control which tabs and functionality are available in the integration detail view.
    """

    # Enables data discovery and monitoring functionality - shows "Data discovery" tab
    DATA_DISCOVERY = "DATA_DISCOVERY"

    # Enables data synchronization to external systems - shows "Data sync" tab
    DATA_SYNC = "DATA_SYNC"

    # Enables task/workflow management for manual processes - shows "Tasks" tab
    TASKS = "TASKS"

    # Indicates integration doesn't require connection testing - shows "Details" tab instead of "Connection" tab
    WITHOUT_CONNECTION = "WITHOUT_CONNECTION"

    # Enables Data Subject Request automation for SAAS integrations
    DSR_AUTOMATION = "DSR_AUTOMATION"

    # Enables conditions configuration for manual task creation - shows "Conditions" tab
    CONDITIONS = "CONDITIONS"



from typing import List, Optional

from pydantic import BaseModel, ConfigDict

from fides.api.schemas.redis_cache import Identity


class SecondPartyResponseFormat(BaseModel):
    """The expected response from a user's HTTP endpoint that receives callbacks from fides.api

    Responses are only expected (and considered) for two_way webhooks.
    """

    derived_identity: Optional[Identity] = None
    halt: bool
    model_config = ConfigDict(use_enum_values=True)


class PrivacyRequestResumeFormat(BaseModel):
    """Expected request body to resume a privacy request after it was paused by a webhook"""

    derived_identity: Optional[Identity] = {}  # type: ignore
    model_config = ConfigDict(use_enum_values=True)


class WebhookJWE(BaseModel):
    """Describes JWE that is given to the user that they need to send with their request
    to resume or pre-approve a privacy request"""

    webhook_id: str
    scopes: List[str]
    iat: str


class RequestTaskJWE(BaseModel):
    """Describes JWE that is given to the user that they need to send to the request
    task callback endpoint to re-queue that request task"""

    request_task_id: str
    scopes: List[str]
    iat: str


class DownloadTokenJWE(BaseModel):
    """Describes JWE that is given to users to access their privacy request download package"""

    privacy_request_id: str
    scopes: List[str]
    iat: str
    exp: str



"""This module contains logic related to loading/manipulation/writing the config."""

import os
from os import environ
from pathlib import Path
from re import compile as regex
from typing import Any, Dict, List, Union

from click import echo
from toml import dump, load

DEFAULT_CONFIG_PATH = ".fides/fides.toml"


def load_file(file_names: Union[List[Path], List[str]]) -> str:
    """
    Load a file from the first matching location.

    In order, will check:
    - A path set at ENV variable FIDES__CONFIG_PATH
    - The current directory
    - The parent directory
    - Two directories up for the current directory
    - The parent_directory/.fides
    - users home (~) directory
    raises FileNotFound if none is found
    """

    possible_directories = [
        os.getenv("FIDES__CONFIG_PATH"),
        os.curdir,
        os.pardir,
        os.path.abspath(os.path.join(os.curdir, "..", "..")),
        os.path.join(os.pardir, ".fides"),
        os.path.expanduser("~"),
    ]

    directories = [d for d in possible_directories if d]

    for dir_str in directories:
        for file_name in file_names:
            possible_location = os.path.join(dir_str, file_name)
            if possible_location and os.path.isfile(possible_location):
                return possible_location

    raise FileNotFoundError


def get_config_from_file(
    config_path_override: str,
    section: str,
    option: str,
) -> Union[str, bool, int, None]:
    """
    Return the value currently written to the config file for the
    specified `section.option`.
    """

    config_path = config_path_override or load_file(file_names=[DEFAULT_CONFIG_PATH])
    current_config = load(config_path)

    config_section = current_config.get(section)
    if config_section is not None:
        config_option = config_section.get(option)
        if config_option is not None:
            return config_option

    return None


def update_config_file(  # type: ignore
    updates: Dict[str, Dict[str, Any]],
    config_path_override: str,
) -> None:
    """
    Overwrite the existing config file with a new version that includes the desired `updates`.

    :param updates: A nested `dict`, where top-level keys correspond to configuration sections and top-level values contain `dict`s whose key/value pairs correspond to the desired option/value updates.
    :param config_path: The path to the current configuration file.
    """

    config_path = config_path_override or load_file(file_names=[DEFAULT_CONFIG_PATH])
    current_config = load(config_path)

    for key, value in updates.items():
        if key in current_config:
            current_config[key].update(value)
        else:
            current_config.update({key: value})

    with open(config_path, "w", encoding="utf-8") as config_file:
        dump(current_config, config_file)

    echo(f"Updated {config_path}:")

    for key, value in updates.items():
        for subkey, val in value.items():
            echo(f"\tSet {key}.{subkey} = {val}")


def handle_deprecated_fields(settings: Dict[str, Any]) -> Dict[str, Any]:
    """Custom logic for handling deprecated values."""

    if settings.get("api") and not settings.get("database"):
        api_settings = settings.pop("api")
        database_settings = {}
        database_settings["user"] = api_settings.get("database_user")
        database_settings["password"] = api_settings.get("database_password")
        database_settings["server"] = api_settings.get("database_host")
        database_settings["port"] = api_settings.get("database_port")
        database_settings["db"] = api_settings.get("database_name")
        database_settings["test_db"] = api_settings.get("test_database_name")
        settings["database"] = database_settings

    return settings


def handle_deprecated_env_variables(settings: Dict[str, Any]) -> Dict[str, Any]:
    """
    Custom logic for handling deprecated ENV variable configuration.
    """

    deprecated_env_vars = regex(r"FIDES__API__(\w+)")

    for key, val in environ.items():
        match = deprecated_env_vars.search(key)
        if match:
            setting = match.group(1).lower()
            setting = setting[setting.startswith("database_") and len("database_") :]
            if setting == "host":
                setting = "server"
            if setting == "name":
                setting = "db"
            if setting == "test_database_name":
                setting = "test_db"

            settings["database"][setting] = val

    return settings



"""adding privacy requeste redaction patterns

Revision ID: 78dbe23d8204
Revises: b1a2c3d4e5f6
Create Date: 2025-08-30 05:40:17.816172

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "78dbe23d8204"
down_revision = "b1a2c3d4e5f6"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "privacy_request_redaction_pattern",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("pattern", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("pattern"),
    )
    op.create_index(
        op.f("ix_privacy_request_redaction_pattern_id"),
        "privacy_request_redaction_pattern",
        ["id"],
        unique=False,
    )


def downgrade():
    op.drop_index(
        op.f("ix_privacy_request_redaction_pattern_id"),
        table_name="privacy_request_redaction_pattern",
    )
    op.drop_table("privacy_request_redaction_pattern")



from typing import Optional

from pydantic import Field, SerializeAsAny
from pydantic_settings import SettingsConfigDict

from fides.api.custom_types import AnyHttpUrlStringRemovesSlash

from .fides_settings import FidesSettings


class PrivacyCenterSettings(FidesSettings):
    """Configuration settings for the Privacy Center."""

    url: SerializeAsAny[Optional[AnyHttpUrlStringRemovesSlash]] = Field(
        default=None, description="The base URL for the Privacy Center."
    )
    model_config = SettingsConfigDict(env_prefix="FIDES__PRIVACY_CENTER__")



from fides.api.api.v1.endpoints import (
    config_endpoints,
    connection_endpoints,
    connection_type_endpoints,
    consent_request_endpoints,
    dataset_config_endpoints,
    drp_endpoints,
    dsr_package_link,
    encryption_endpoints,
    identity_verification_endpoints,
    manual_webhook_endpoints,
    masking_endpoints,
    messaging_endpoints,
    oauth_endpoints,
    partitioning_endpoints,
    policy_endpoints,
    policy_webhook_endpoints,
    pre_approval_webhook_endpoints,
    privacy_request_endpoints,
    privacy_request_redaction_patterns_endpoints,
    registration_endpoints,
    saas_config_endpoints,
    storage_endpoints,
    user_endpoints,
    user_permission_endpoints,
    worker_endpoints,
)
from fides.api.util.api_router import APIRouter

api_router = APIRouter()
api_router.include_router(config_endpoints.router)
api_router.include_router(connection_type_endpoints.router)
api_router.include_router(connection_endpoints.router)
api_router.include_router(consent_request_endpoints.router)
api_router.include_router(dataset_config_endpoints.router)
api_router.include_router(drp_endpoints.router)
api_router.include_router(dsr_package_link.router)
api_router.include_router(encryption_endpoints.router)
api_router.include_router(masking_endpoints.router)
api_router.include_router(oauth_endpoints.router)
api_router.include_router(policy_endpoints.router)
api_router.include_router(policy_webhook_endpoints.router)
api_router.include_router(pre_approval_webhook_endpoints.router)
api_router.include_router(privacy_request_endpoints.router)
api_router.include_router(privacy_request_redaction_patterns_endpoints.router)
api_router.include_router(identity_verification_endpoints.router)
api_router.include_router(storage_endpoints.router)
api_router.include_router(messaging_endpoints.router)
api_router.include_router(saas_config_endpoints.router)
api_router.include_router(partitioning_endpoints.router)
api_router.include_router(user_endpoints.router)
api_router.include_router(user_permission_endpoints.router)
api_router.include_router(manual_webhook_endpoints.router)
api_router.include_router(registration_endpoints.router)
api_router.include_router(worker_endpoints.router)



"""Base abstract class for cloud storage clients."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Optional

from fideslang.validation import AnyHttpUrlString


class BaseStorageClient(ABC):
    """Abstract base class for cloud storage clients.

    This class defines the interface that all provider-specific storage clients
    must implement. It provides a common contract for URI building, transport
    parameters, and presigned URL generation.
    """

    def __init__(self, storage_secrets: Any):
        """Initialize the storage client with secrets.

        Args:
            storage_secrets: Provider-specific storage credentials and configuration.
                           Derived classes will specify the exact type they expect.
        """
        self.storage_secrets = storage_secrets

    @abstractmethod
    def build_uri(self, bucket: str, key: str) -> str:
        """Build the URI for the storage location.

        Args:
            bucket: Storage bucket/container name
            key: Object key/path

        Returns:
            URI string for smart-open
        """

    @abstractmethod
    def get_transport_params(self) -> dict[str, Any]:
        """Get transport parameters for smart-open.

        Returns:
            Dictionary of transport parameters for smart-open
        """

    @abstractmethod
    def generate_presigned_url(
        self, bucket: str, key: str, ttl_seconds: Optional[int] = None
    ) -> AnyHttpUrlString:
        """Generate a presigned URL for the object.

        Args:
            bucket: Storage bucket/container name
            key: Object key/path
            ttl_seconds: Time to live in seconds

        Returns:
            Presigned URL for the object
        """



from typing import Annotated, List

from fastapi import Body, Depends, Security
from fastapi_pagination import Page, Params
from fastapi_pagination.bases import AbstractPage
from fastapi_pagination.ext.sqlalchemy import paginate
from fideslang.validation import FidesKey
from loguru import logger
from pydantic import Field
from sqlalchemy.orm import Session
from starlette.exceptions import HTTPException
from starlette.status import HTTP_200_OK, HTTP_400_BAD_REQUEST, HTTP_404_NOT_FOUND

from fides.api.api import deps
from fides.api.common_exceptions import KeyOrNameAlreadyExists
from fides.api.db.base_class import get_key_from_data
from fides.api.models.pre_approval_webhook import PreApprovalWebhook
from fides.api.oauth.utils import verify_oauth_client
from fides.api.schemas import pre_approval_webhooks as schemas
from fides.api.util.api_router import APIRouter
from fides.api.util.connection_util import get_connection_config_or_error
from fides.common.api import scope_registry as scopes
from fides.common.api.v1 import urn_registry as urls

router = APIRouter(tags=["Pre Approval Webhooks"], prefix=urls.V1_URL_PREFIX)


@router.get(
    urls.WEBHOOK_PRE_APPROVAL,
    status_code=HTTP_200_OK,
    response_model=Page[schemas.PreApprovalWebhookResponse],
    dependencies=[Security(verify_oauth_client, scopes=[scopes.WEBHOOK_READ])],
)
def get_pre_approval_webhook_list(
    *,
    db: Session = Depends(deps.get_db),
    params: Params = Depends(),
) -> AbstractPage[schemas.PreApprovalWebhookResponse]:
    """
    Return a paginated list of all PreApprovalWebhook records in this system
    """
    logger.debug(
        "Finding all pre_approval webhooks with pagination params '{}'", params
    )
    pre_approval_webhooks = PreApprovalWebhook.query(db=db).order_by(
        PreApprovalWebhook.created_at.desc()
    )
    return paginate(pre_approval_webhooks, params=params)


def get_pre_approval_webhook_or_error(
    db: Session, webhook_key: FidesKey
) -> PreApprovalWebhook:
    """Helper method to load PreApprovalWebhook or throw a 404"""
    logger.debug("Finding PreApprovalWebhook with key '{}'", webhook_key)
    pre_approval_webhook = PreApprovalWebhook.get_by(
        db=db, field="key", value=webhook_key
    )
    if not pre_approval_webhook:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No PreApprovalWebhook found for key {webhook_key}.",
        )

    return pre_approval_webhook


@router.get(
    urls.WEBHOOK_PRE_APPROVAL_DETAIL,
    status_code=HTTP_200_OK,
    response_model=schemas.PreApprovalWebhookResponse,
    dependencies=[Security(verify_oauth_client, scopes=[scopes.WEBHOOK_READ])],
)
def get_pre_approval_webhook_detail(
    *,
    db: Session = Depends(deps.get_db),
    webhook_key: FidesKey,
) -> PreApprovalWebhook:
    """
    Loads the given Pre-Approval Webhook by key
    """
    return get_pre_approval_webhook_or_error(db, webhook_key)


@router.put(
    urls.WEBHOOK_PRE_APPROVAL,
    status_code=HTTP_200_OK,
    dependencies=[
        Security(verify_oauth_client, scopes=[scopes.WEBHOOK_CREATE_OR_UPDATE])
    ],
    response_model=List[schemas.PreApprovalWebhookResponse],
)
def create_or_update_pre_execution_webhooks(
    *,
    db: Session = Depends(deps.get_db),
    webhooks: Annotated[List[schemas.PreApprovalWebhookCreate], Field(max_length=50)],  # type: ignore
) -> List[PreApprovalWebhook]:
    """
    Create or update the list of Pre-Approval Webhooks that run as soon as a request is created.
    This endpoint is all-or-nothing.
    Either all webhooks should be created/updated or none should be updated. Deletes any webhooks not present
    in the request body.
    """
    pre_approval_webhooks = PreApprovalWebhook.query(db=db)

    keys = [
        get_key_from_data(webhook.model_dump(mode="json"), PreApprovalWebhook.__name__)
        for webhook in webhooks
    ]
    names = [webhook.name for webhook in webhooks]
    if len(keys) != len(set(keys)) or len(names) != len(set(names)):
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail="Check request body: there are multiple webhooks whose keys or names resolve to the same value.",
        )

    staged_webhooks = []  # Webhooks will be committed at the end
    for schema in webhooks:
        connection_config = get_connection_config_or_error(
            db, schema.connection_config_key
        )
        try:
            webhook = PreApprovalWebhook.create_or_update(
                db=db,
                data={
                    "key": schema.key,
                    "name": schema.name,
                    "connection_config_id": connection_config.id,
                },
            )
            staged_webhooks.append(webhook)
        except KeyOrNameAlreadyExists as exc:
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail=exc.args[0],
            )

    staged_webhook_keys = [webhook.key for webhook in staged_webhooks]
    webhooks_to_remove = pre_approval_webhooks.filter(
        PreApprovalWebhook.key.not_in(staged_webhook_keys)  # type: ignore
    )

    if webhooks_to_remove.count():
        logger.info(
            "Removing Pre-Approval Webhooks that were not included in request: {}",
            [webhook.key for webhook in webhooks_to_remove],
        )
        webhooks_to_remove.delete()

    logger.info(
        "Creating/updating Policy Pre-Execution Webhooks: {}", staged_webhook_keys
    )
    # Committing to database now, as a last step, once we've verified that all the webhooks
    # in the request are free of issues.
    db.commit()
    return staged_webhooks


@router.patch(
    urls.WEBHOOK_PRE_APPROVAL_DETAIL,
    status_code=HTTP_200_OK,
    dependencies=[
        Security(verify_oauth_client, scopes=[scopes.WEBHOOK_CREATE_OR_UPDATE])
    ],
    response_model=schemas.PreApprovalWebhookResponse,
)
def update_pre_execution_webhook(
    *,
    db: Session = Depends(deps.get_db),
    webhook_key: FidesKey,
    webhook_body: schemas.PreApprovalWebhookUpdate = Body(...),
) -> schemas.PreApprovalWebhookResponse:
    """PATCH a single Pre-Approval Webhook that runs as soon as Privacy Request is created."""

    loaded_webhook = get_pre_approval_webhook_or_error(db, webhook_key)
    data = webhook_body.model_dump(exclude_none=True)

    if data.get("connection_config_key"):
        connection_config = get_connection_config_or_error(
            db, data.get("connection_config_key")  # type: ignore
        )
        data["connection_config_id"] = connection_config.id

    try:
        logger.info(
            "Updating Pre-Approval Webhook with key '{}' ",
            webhook_key,
        )
        loaded_webhook.update(db, data=data)
    except KeyOrNameAlreadyExists as exc:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=exc.args[0],
        )

    # Pre Execution Webhooks are not committed by default, so we commit at the end.
    db.commit()
    return schemas.PreApprovalWebhookResponse(
        connection_config=loaded_webhook.connection_config,
        key=loaded_webhook.key,
        name=loaded_webhook.name,
    )


@router.delete(
    urls.WEBHOOK_PRE_APPROVAL_DETAIL,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[scopes.WEBHOOK_DELETE])],
)
def delete_pre_execution_webhook(
    *,
    db: Session = Depends(deps.get_db),
    webhook_key: FidesKey,
) -> None:
    """Delete the Pre-Approval Webhook given webhook key"""
    loaded_webhook = get_pre_approval_webhook_or_error(db, webhook_key)

    logger.info(
        "Deleting Pre-Approval Webhook with key '{}'",
        webhook_key,
    )
    loaded_webhook.delete(db=db)



from typing import Any, Optional, Union

from loguru import logger
from sqlalchemy.orm import Session

from fides.api.graph.config import FieldPath
from fides.api.task.conditional_dependencies.operators import (
    LOGICAL_OPERATORS,
    OPERATOR_METHODS,
)
from fides.api.task.conditional_dependencies.schemas import (
    Condition,
    ConditionEvaluationResult,
    ConditionGroup,
    ConditionLeaf,
    EvaluationResult,
    GroupEvaluationResult,
    Operator,
)


class ConditionEvaluationError(Exception):
    """Error raised when a condition evaluation fails"""


class ConditionEvaluator:
    """Evaluates nested conditions and returns a boolean result and a detailed evaluation report"""

    def __init__(self, db: Session):
        self.db = db

    def evaluate_rule(
        self, rule: Condition, data: Union[dict, Any]
    ) -> EvaluationResult:
        """Evaluate a nested condition rule against input data and return detailed results

        Args:
            rule: The condition rule to evaluate
            data: The data to evaluate the condition against

        Returns:
            evaluation report: A detailed report of the evaluation
            - The field address of the condition
            - The operator used in the condition
            - The expected value of the condition
            - The actual value of the condition
            - The result of the condition evaluation
            - A message describing the condition evaluation
        """
        if isinstance(rule, ConditionLeaf):
            leaf_result = self._evaluate_leaf_condition(rule, data)
            return leaf_result
        # ConditionGroup
        group_result = self._evaluate_group_condition(rule, data)
        return group_result

    def _evaluate_leaf_condition(
        self, condition: ConditionLeaf, data: Union[dict, Any]
    ) -> ConditionEvaluationResult:
        """Evaluate a leaf condition against input data

        Args:
            condition: The leaf condition to evaluate
            data: The data to evaluate the condition against

        Returns:
            A detailed evaluation report for the leaf condition

        Raises:
            ConditionEvaluationError: If there is an issue applying the operator or if an unexpected error occurs.
        """
        # Handle both colon-separated and dot-separated field addresses
        if ":" in condition.field_address:
            # Full field address like "dataset:collection:field" - split on colons
            keys = condition.field_address.split(":")
        else:
            # Relative field path like "field.subfield" - split on dots
            keys = condition.field_address.split(".")

        data_value = self._get_nested_value(data, keys)

        # Apply operator and get result
        try:
            result = self._apply_operator(
                data_value, condition.operator, condition.value
            )
            message = f"Condition '{condition.field_address} {condition.operator} {condition.value}' evaluated to {result}"
        except ConditionEvaluationError as e:
            logger.error(
                f"Unexpected error evaluating condition '{condition.field_address} {condition.operator} {condition.value}': {str(e)}"
            )
            raise

        return ConditionEvaluationResult(
            field_address=condition.field_address,
            operator=condition.operator,
            expected_value=condition.value,
            actual_value=data_value,
            result=result,
            message=message,
        )

    def _evaluate_group_condition(
        self, group: ConditionGroup, data: Union[dict, Any]
    ) -> GroupEvaluationResult:
        """Evaluate a group condition against input data

        Args:
            group: The group condition to evaluate
            data: The data to evaluate the condition against

        Returns:
            A detailed evaluation report for the group condition

        Raises:
            ConditionEvaluationError: If there is an issue evaluating the group condition (e.g., from evaluate_rule calls)
        """
        try:
            operator_func = LOGICAL_OPERATORS[group.logical_operator]
        except KeyError as e:
            raise ConditionEvaluationError(
                f"Unknown logical operator: {group.logical_operator}"
            ) from e

        results = [
            self.evaluate_rule(condition, data) for condition in group.conditions
        ]
        group_result = operator_func([r.result for r in results])

        return GroupEvaluationResult(
            logical_operator=group.logical_operator,
            condition_results=results,
            result=group_result,
        )

    def _get_nested_value_from_fides_reference_structure(
        self, data: Any, keys: list[str]
    ) -> Optional[Any]:
        """Get nested value from Fides reference structure

        Args:
            data: The Fides reference structure to get the nested value from
            keys: The keys to for the specific nested value in the data

        Returns:
            The nested value from the data or None if not a Fides reference structure

        Raises:
            AttributeError: If the data does not have a get_field_value method
            ValueError: If the keys are not valid for the Fides reference structure
        """
        if hasattr(data, "get_field_value"):
            try:
                field_path = FieldPath(*keys) if len(keys) > 1 else FieldPath(keys[0])
                return data.get_field_value(field_path)
            except (AttributeError, ValueError):
                logger.debug(
                    f"Fides reference structure does not have a get_field_value method: {data}"
                )
                raise
        raise ConditionEvaluationError(
            f"Data does not have a get_field_value method: {data}"
        )

    def _get_nested_value_from_dict(self, data: dict, keys: list[str]) -> Optional[Any]:
        """Get nested value from dictionary. This is the fallback and will return None if the key is not found.
        When the data is missing the None value will work with exists/not_exists operations and correctly evaluate to False
        for other operations like eq, not_eq, etc.

        Args:
            data: The dictionary to get the nested value from
            keys: The keys to for the specific nested value in the data

        Returns:
            The nested value from the data

        Raises:
            KeyError: If the keys are not valid for the dictionary
        """
        current: Any = data
        for key in keys:
            if not isinstance(current, dict):
                return None
            current = current.get(key)
            if current is None:
                return None
        return current

    def _get_nested_value(self, data: Union[dict, Any], keys: list[str]) -> Any:
        """Get nested value from data using dot notation or colon notation

        Supports both simple dictionary access and Fides reference structures:
        - Simple dict: data["user"]["name"]
        - Fides FieldAddress: data.get_field_value(FieldAddress("dataset", "collection", "field_address"))
        - Fides Collection: data.get_field_value(FieldPath("field_address", "subfield"))

        Also supports full field addresses with dataset:collection:field format

        Args:
            data: The data to get the nested value from
            keys: The keys to for the specific nested value in the data

        Returns:
            The nested value from the data

        Raises:
            KeyError: If the keys are not valid for the dictionary
        """
        if not keys:
            return data

        # Try Fides reference structures first
        try:
            return self._get_nested_value_from_fides_reference_structure(data, keys)
        except (AttributeError, ValueError, ConditionEvaluationError):
            pass

        # Fall back to dictionary access for all path types
        return self._get_nested_value_from_dict(data, keys)

    def _apply_operator(
        self, data_value: Any, operator: Operator, user_input_value: Any
    ) -> bool:
        """Apply operator to actual and expected values
        The operator is validated in the ConditionLeaf and ConditionGroup schemas,
        so we don't need to validate it here.

        Args:
            data_value: The actual value to evaluate
            operator: The operator to apply
            user_input_value: The expected value to evaluate against

        Returns:
            The result of the operator applied to the actual and expected values
        """
        # Get the method for the operator and execute it
        try:
            operator_method = OPERATOR_METHODS[operator]
            return operator_method(data_value, user_input_value)
        except KeyError as e:
            # Unknown operator
            logger.error(f"Unknown operator: {operator}")
            raise ConditionEvaluationError(f"Unknown operator: {operator}") from e
        except Exception as e:
            # Log unexpected errors but still raise them
            logger.error(f"Unexpected error in operator {operator}: {e}")
            raise ConditionEvaluationError(
                f"Unexpected error evaluating condition: {e}"
            ) from e



from __future__ import annotations

from io import BytesIO
from typing import Any, Optional, Union

from fideslang.validation import AnyHttpUrlString
from loguru import logger

from fides.api.common_exceptions import StorageUploadError
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.schemas.storage.storage import StorageSecrets, StorageSecretsS3
from fides.api.service.storage.s3 import generic_upload_to_s3
from fides.api.service.storage.streaming.retry import retry_cloud_storage_operation
from fides.api.service.storage.streaming.schemas import StorageUploadConfig
from fides.api.service.storage.streaming.smart_open_client import SmartOpenStorageClient
from fides.api.service.storage.streaming.smart_open_streaming_storage import (
    SmartOpenStreamingStorage,
)


@retry_cloud_storage_operation(
    provider="s3_streaming",
    operation_name="upload_to_s3_streaming",
    max_retries=2,
    base_delay=2.0,
    max_delay=30.0,
)
def upload_to_s3_streaming(
    storage_secrets: Union[StorageSecretsS3, dict[StorageSecrets, Any]],
    data: dict,
    bucket_name: str,
    file_key: str,
    resp_format: str,
    privacy_request: PrivacyRequest,
    document: Optional[BytesIO],
    auth_method: str,
    max_workers: int = 5,
) -> Optional[AnyHttpUrlString]:
    """Uploads arbitrary data to S3 using smart-open streaming for memory efficiency.

    This function now uses smart-open for efficient cloud storage operations while maintaining
    our DSR-specific business logic for package splitting and attachment processing.
    """
    formatted_secrets = format_secrets(storage_secrets)

    if document is not None:
        _, response = generic_upload_to_s3(
            formatted_secrets,  # type: ignore[arg-type]
            bucket_name,
            file_key,
            auth_method,
            document,
        )
        return response

    try:
        logger.debug("Creating SmartOpenStorageClient with formatted secrets")
        storage_client = SmartOpenStorageClient("s3", auth_method, formatted_secrets)

        # Create upload config for the streaming interface
        upload_config = StorageUploadConfig(
            bucket_name=bucket_name,
            file_key=file_key,
            resp_format=resp_format,
            max_workers=max_workers,
        )

        # Use the smart-open streaming implementation
        streaming_storage = SmartOpenStreamingStorage(storage_client)
        result = streaming_storage.upload_to_storage_streaming(
            data,
            upload_config,
            privacy_request,
            document,
        )

        logger.debug(
            "Successfully uploaded streaming archive to S3 using smart-open: {}",
            file_key,
        )
        return result

    except Exception as e:
        logger.error("Unexpected error during smart-open streaming upload: {}", e)
        raise StorageUploadError(
            f"Unexpected error during smart-open streaming upload: {e}"
        )


def _process_storage_secrets_input(
    storage_secrets: Union[StorageSecretsS3, dict[StorageSecrets, Any]]  # type: ignore[misc]
) -> dict[str, Any]:
    """Process input and convert to string-keyed dictionary."""
    final_secrets: dict[str, Any] = {}

    logger.debug(
        f"Processing storage secrets input of type: {type(storage_secrets).__name__}"
    )

    if isinstance(storage_secrets, StorageSecretsS3):
        # Convert StorageSecretsS3 model directly to string keys
        for key, value in storage_secrets.model_dump().items():
            if value is not None:
                final_secrets[key] = value
    else:
        # Process dict input, converting enum keys to strings if needed
        for key, value in (storage_secrets or {}).items():  # type: ignore[assignment]
            if isinstance(key, str):
                final_secrets[key] = value
            elif isinstance(key, StorageSecrets):
                final_secrets[key.value] = value

    return final_secrets


def _validate_aws_credentials(final_secrets: dict[str, Any]) -> None:
    """Validate that required AWS credentials are present."""
    has_access_key = (
        "aws_access_key_id" in final_secrets and final_secrets["aws_access_key_id"]
    )
    has_secret_key = (
        "aws_secret_access_key" in final_secrets
        and final_secrets["aws_secret_access_key"]
    )

    # If we have any AWS credentials, we need both for SECRET_KEYS auth
    if has_access_key or has_secret_key:
        if not has_access_key:
            raise ValueError(
                "Missing required AWS credentials for SECRET_KEYS auth: aws_access_key_id. "
                "Storage configuration must include valid AWS access key and secret key."
            )
        if not has_secret_key:
            raise ValueError(
                "Missing required AWS credentials for SECRET_KEYS auth: aws_secret_access_key. "
                "Storage configuration must include valid AWS access key and secret key."
            )
    else:
        # AUTOMATIC authentication - check if region is provided
        has_region = "region_name" in final_secrets and final_secrets["region_name"]

        if not has_region:
            raise ValueError(
                "Missing required region_name for AUTOMATIC authentication. "
                "Storage configuration must include a valid AWS region."
            )


def format_secrets(
    storage_secrets: Union[StorageSecretsS3, dict[StorageSecrets, Any]]  # type: ignore[misc]
) -> dict[str, Any]:
    """
    Returns the correct format for the S3StorageClient.

    This function handles multiple credential formats and processes them through several stages:
    1. Input processing: Accepts either StorageSecretsS3 models (from API) or raw dicts (from database)
    2. Key normalization: Converts all keys to string format for consistency
    3. Default setting: Sets default region if missing (before validation for better automatic auth support)
    4. Validation: Ensures required AWS credentials are present based on auth method
    5. Return: Returns string-keyed dict ready for S3StorageClient

    Input formats:
    - StorageSecretsS3: Used by storage API endpoints (e.g., put_config_secrets)
    - dict[StorageSecrets, Any]: Used by storage models (e.g., StorageConfig.secrets)
    - dict[str, Any]: Used by database queries (e.g., StorageConfig.get_by)

    Authentication methods:
    - SECRET_KEYS: Requires aws_access_key_id, aws_secret_access_key, and region_name
    - AUTOMATIC: Requires only region_name (relies on AWS SDK credential chain)
    - Role assumption: Can be used with either auth method via assume_role_arn

    Output format:
    - dict[str, Any]: Required by S3StorageClient.generate_presigned_url() and other AWS operations

    Returns:
        dict[str, Any]: Credentials with string keys like 'aws_access_key_id'

    Raises:
        ValueError: If required AWS credentials are missing for the chosen auth method
    """
    final_secrets = _process_storage_secrets_input(storage_secrets)
    _validate_aws_credentials(final_secrets)

    return final_secrets



from typing import Any, Dict, Optional

from fideslang.validation import FidesKey
from pydantic import BaseModel, ConfigDict

from fides.api.models.connectionconfig import AccessLevel, ConnectionType
from fides.api.schemas.connection_configuration import connection_secrets_schemas


class SaasConnectionTemplateValues(BaseModel):
    """Schema with values to create both a Saas ConnectionConfig and DatasetConfig from a template"""

    name: Optional[str] = None  # For ConnectionConfig
    key: Optional[FidesKey] = None  # For ConnectionConfig
    description: Optional[str] = None  # For ConnectionConfig
    secrets: connection_secrets_schemas  # For ConnectionConfig
    instance_key: FidesKey  # For DatasetConfig.fides_key
    model_config = ConfigDict(extra="ignore")

    def generate_config_data_from_template(
        self, config_from_template: Dict
    ) -> Dict[str, Any]:
        """Generate a config data object (dict) based on the template values"""
        data = {
            "key": self.key if self.key else self.instance_key,
            "description": self.description,
            "connection_type": ConnectionType.saas,
            "access": AccessLevel.write,  # does this need to be passed as a method arg instead?
            "saas_config": config_from_template,
        }
        if self.name:
            data["name"] = self.name
        return data



import re

from anyascii import anyascii  # type: ignore

# Regex to validate a ISO 3166-2 code:
# 1. Starts with a 2 letter country code (e.g. "US", "GB") (see ISO 3166-1 alpha-2)
# 2. (Optional) Ends with a 1-3 alphanumeric character region code (e.g. "CA", "123", "X") (see ISO 3166-2)
# 3. Country & region codes must be separated by a hyphen (e.g. "US-CA")
#
# Fides also supports a special `EEA` geolocation code to denote the European
# Economic Area; this is not part of ISO 3166-2, but is supported for convenience.
VALID_ISO_3166_LOCATION_REGEX = re.compile(
    r"^(?:([a-z]{2})(-[a-z0-9]{1,3})?|(eea))$", re.IGNORECASE
)


def normalize_location_code(location: str) -> str:
    """
    Normalize a location code to ISO 3166-2 format.

    Converts location codes to the idiomatic format:
    - Uppercase (e.g., "us-ca" -> "US-CA")
    - Hyphens instead of underscores (e.g., "US_CA" -> "US-CA")
    - Supports special codes like "EEA"

    Args:
        location: Input location code (e.g., "us-ca", "US_NY", "gb", "eea")

    Returns:
        Normalized location code (e.g., "US-CA", "US-NY", "GB", "EEA")

    Raises:
        ValueError: If the normalized location doesn't match ISO 3166-2 format

    Examples:
        normalize_location_code("us-ca") -> "US-CA"
        normalize_location_code("US_NY") -> "US-NY"
        normalize_location_code("gb") -> "GB"
        normalize_location_code("eea") -> "EEA"
    """
    if not location:
        raise ValueError("Location code cannot be empty")

    # Convert to uppercase and replace underscores with hyphens
    normalized = location.upper().replace("_", "-")

    # Validate against ISO 3166-2 format
    if not VALID_ISO_3166_LOCATION_REGEX.match(normalized):
        raise ValueError(
            f"Invalid location format '{location}'. Must follow ISO 3166 format "
            "(e.g., 'US', 'US-CA', 'GB', 'CA-ON', 'EEA')"
        )

    return normalized


def to_snake_case(text: str) -> str:
    """
    Returns a snake-cased str based upon the input text.

    Examples:
        "foo bar" -> "foo_bar"
        "foo\nbar" -> "foo_bar"
        "foo\tbar" -> "foo_bar"
        "foo-bar" -> "foo_bar"
        "foo*bar" -> "foobar"
    """
    text = anyascii(text).lower().strip()
    text = re.sub(r"[^\w\s-]", "", text)
    text = re.sub(r"[\s-]+", "_", text)
    text = re.sub(r"^-+|-+$", "", text)
    return text



from __future__ import annotations

from abc import ABC
from typing import Any, Dict, Generic, List, Optional, Type, TypeVar

from pydantic import ValidationError

from fides.api.common_exceptions import NoSuchStrategyException
from fides.api.common_exceptions import ValidationError as FidesopsValidationError
from fides.api.schemas.saas.strategy_configuration import StrategyConfiguration

T = TypeVar("T", bound="Strategy")
C = TypeVar("C", bound=StrategyConfiguration)


def _find_strategy_subclass(
    cls: Type[Strategy], strategy_name: str
) -> Optional[Type[Strategy]]:
    if hasattr(cls, "name") and cls.name == strategy_name:
        return cls
    for sub in cls.__subclasses__():
        found = _find_strategy_subclass(sub, strategy_name)
        if found:
            return found
    return None


def _find_all_strategy_subclasses(
    cls: Type[T], subs: Optional[List[Type[T]]] = None
) -> List[Type[T]]:
    if subs is None:
        subs = []
    if hasattr(cls, "name"):
        subs.append(cls)
    for sub in cls.__subclasses__():
        _find_all_strategy_subclasses(sub, subs)
    return subs


class Strategy(ABC, Generic[C]):
    """Abstract base class for strategies"""

    name: str
    configuration_model: Optional[Type[C]] = None

    @classmethod
    def get_strategy(
        cls: Type[T],
        strategy_name: str,
        configuration: Optional[Dict[str, Any]] = None,
    ) -> T:
        """
        Returns the strategy given the name and configuration.
        Raises NoSuchStrategyException if the strategy does not exist
        """

        strategy_class = _find_strategy_subclass(cls, strategy_name)

        if strategy_class is None:
            valid_strategies = ", ".join(
                [sub.name for sub in _find_all_strategy_subclasses(cls)]
            )
            raise NoSuchStrategyException(
                f"Strategy '{strategy_name}' does not exist. Valid strategies are [{valid_strategies}]"
            )
        try:
            if strategy_class.configuration_model:
                strategy_config = strategy_class.configuration_model(**configuration)
            else:
                return strategy_class()  # type: ignore
        except ValidationError as e:
            raise FidesopsValidationError(message=str(e))
        return strategy_class(strategy_config)  # type: ignore

    @classmethod
    def get_strategies(cls: Type[T]) -> List[Type[T]]:
        """Returns all supported strategies"""
        return _find_all_strategy_subclasses(cls)



from fastapi import Depends, HTTPException, Security
from fastapi.routing import APIRouter
from loguru import logger
from sqlalchemy.orm import Session
from starlette.status import HTTP_200_OK

from fides.api.api.deps import get_db
from fides.api.models.privacy_request_redaction_pattern import (
    PrivacyRequestRedactionPattern,
)
from fides.api.oauth.utils import verify_oauth_client
from fides.api.schemas.privacy_request_redaction_patterns import (
    PrivacyRequestRedactionPatternsRequest,
    PrivacyRequestRedactionPatternsResponse,
)
from fides.common.api.scope_registry import (
    PRIVACY_REQUEST_REDACTION_PATTERNS_READ,
    PRIVACY_REQUEST_REDACTION_PATTERNS_UPDATE,
)
from fides.common.api.v1.urn_registry import (
    PRIVACY_REQUEST_REDACTION_PATTERNS,
    V1_URL_PREFIX,
)

router = APIRouter(
    tags=["Privacy Request Redaction Patterns"],
    prefix=V1_URL_PREFIX,
)


@router.get(
    PRIVACY_REQUEST_REDACTION_PATTERNS,
    status_code=HTTP_200_OK,
    response_model=PrivacyRequestRedactionPatternsResponse,
    dependencies=[
        Security(verify_oauth_client, scopes=[PRIVACY_REQUEST_REDACTION_PATTERNS_READ])
    ],
)
def get_privacy_request_redaction_patterns(
    *, db: Session = Depends(get_db)
) -> PrivacyRequestRedactionPatternsResponse:
    """
    Get the current privacy request redaction patterns configuration.

    Returns the list of regex patterns used to mask dataset, collection,
    and field names in privacy request package reports.
    """
    logger.info("Getting privacy request redaction patterns configuration")

    patterns = PrivacyRequestRedactionPattern.get_patterns(db)
    return PrivacyRequestRedactionPatternsResponse(patterns=patterns)


@router.put(
    PRIVACY_REQUEST_REDACTION_PATTERNS,
    status_code=HTTP_200_OK,
    response_model=PrivacyRequestRedactionPatternsResponse,
    dependencies=[
        Security(
            verify_oauth_client, scopes=[PRIVACY_REQUEST_REDACTION_PATTERNS_UPDATE]
        )
    ],
)
def update_privacy_request_redaction_patterns(
    *,
    db: Session = Depends(get_db),
    request: PrivacyRequestRedactionPatternsRequest,
) -> PrivacyRequestRedactionPatternsResponse:
    """
    Update the privacy request redaction patterns configuration.

    This is a complete replacement of the patterns list. To clear all patterns,
    send an empty list.
    """
    logger.info(
        "Updating privacy request redaction patterns configuration with {} patterns",
        len(request.patterns),
    )

    try:
        updated = PrivacyRequestRedactionPattern.replace_patterns(
            db=db, patterns=request.patterns
        )

        logger.info("Successfully updated privacy request redaction patterns")
        return PrivacyRequestRedactionPatternsResponse(patterns=updated)

    except Exception as exc:
        logger.exception(
            "Failed to update privacy request redaction patterns: {}", str(exc)
        )
        raise HTTPException(
            status_code=500,
            detail="Failed to update privacy request redaction patterns",
        ) from exc



"""
Provides functions that seed the application with data.
"""

import os
from typing import Dict, List, Optional

from fideslang.default_taxonomy import DEFAULT_TAXONOMY
from loguru import logger as log
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session

from fides.api.api.v1.endpoints.dataset_config_endpoints import patch_dataset_configs
from fides.api.api.v1.endpoints.saas_config_endpoints import (
    instantiate_connection_from_template,
)
from fides.api.common_exceptions import KeyOrNameAlreadyExists
from fides.api.db.base_class import FidesBase
from fides.api.db.ctl_session import sync_session
from fides.api.db.system import upsert_system
from fides.api.models.application_config import ApplicationConfig
from fides.api.models.client import ClientDetail
from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.models.datasetconfig import DatasetConfig
from fides.api.models.fides_user import FidesUser
from fides.api.models.fides_user_permissions import FidesUserPermissions
from fides.api.models.policy import Policy, Rule, RuleTarget
from fides.api.models.sql_models import (  # type: ignore[attr-defined]
    Dataset,
    Organization,
    System,
    sql_model_map,
)
from fides.api.oauth.roles import OWNER
from fides.api.schemas.connection_configuration.connection_config import (
    CreateConnectionConfigurationWithSecrets,
)
from fides.api.schemas.connection_configuration.saas_config_template_values import (
    SaasConnectionTemplateValues,
)
from fides.api.schemas.dataset import DatasetConfigCtlDataset
from fides.api.schemas.messaging.messaging import (
    MessagingConfigRequest,
    MessagingServiceDetailsMailgun,
    MessagingServiceType,
)
from fides.api.schemas.messaging.shared_schemas import MessagingSecretsMailgunDocs
from fides.api.schemas.policy import ActionType, DrpAction
from fides.api.service.messaging.messaging_crud_service import (
    create_or_update_messaging_config,
)
from fides.api.util.connection_util import patch_connection_configs
from fides.api.util.data_category import get_user_data_categories
from fides.api.util.errors import AlreadyExistsError, QueryError
from fides.api.util.text import to_snake_case
from fides.config import CONFIG
from fides.service.dataset.dataset_config_service import DatasetConfigService

from .crud import upsert_resources
from .samples import (
    load_sample_connections_from_project,
    load_sample_resources_from_project,
)

DEFAULT_OAUTH_CLIENT_KEY = "default_oauth_client"
DEFAULT_ACCESS_POLICY = "default_access_policy"
DEFAULT_ACCESS_POLICY_RULE = "default_access_policy_rule"
DEFAULT_ERASURE_POLICY = "default_erasure_policy"
DEFAULT_ERASURE_POLICY_RULE = "default_erasure_policy_rule"
DEFAULT_ERASURE_MASKING_STRATEGY = "hmac"

DEFAULT_CONSENT_POLICY: str = "default_consent_policy"
DEFAULT_CONSENT_RULE = "default_consent_rule"


def create_or_update_parent_user() -> None:
    with sync_session() as db_session:
        if (
            not CONFIG.security.parent_server_username
            and not CONFIG.security.parent_server_password
        ):
            return

        if (
            CONFIG.security.parent_server_username
            and not CONFIG.security.parent_server_password
            or CONFIG.security.parent_server_password
            and not CONFIG.security.parent_server_username
        ):
            # Both log and raise are here because the raise message is not showing.
            # It could potentially be related to https://github.com/ethyca/fides/issues/1228
            log.error(
                "Both a parent_server_user and parent_server_password must be set to create a parent server user"
            )
            raise ValueError(
                "Both a parent_server_user and parent_server_password must be set to create a parent server user"
            )

        user = (
            FidesUser.get_by(
                db_session,
                field="username",
                value=CONFIG.security.parent_server_username,
            )
            if CONFIG.security.parent_server_username
            else None
        )

        if user and CONFIG.security.parent_server_password:
            if not user.credentials_valid(CONFIG.security.parent_server_password):
                log.debug("Updating Fides parent user credentials")
                user.update_password(db_session, CONFIG.security.parent_server_password)
                return
            # clean exit if parent user already exists and credentials match
            return

        log.debug("Creating Fides parent user credentials")
        user = FidesUser.create(
            db=db_session,
            data={
                "username": CONFIG.security.parent_server_username,
                "password": CONFIG.security.parent_server_password,
            },
        )
        FidesUserPermissions.create(
            db=db_session,
            data={"user_id": user.id, "roles": [OWNER]},
        )


def get_client_id(db_session: Session) -> str:
    client = ClientDetail.get_by(
        db=db_session,
        field="fides_key",
        value=DEFAULT_OAUTH_CLIENT_KEY,
    )
    if not client:
        client, _ = ClientDetail.create_client_and_secret(
            db=db_session,
            client_id_byte_length=CONFIG.security.oauth_client_id_length_bytes,
            client_secret_byte_length=CONFIG.security.oauth_client_secret_length_bytes,
            fides_key=DEFAULT_OAUTH_CLIENT_KEY,
        )

    return client.id


def load_default_access_policy(
    db_session: Session, client_id: str, default_data_categories: List[str]
) -> None:
    access_policy: Optional[FidesBase] = Policy.get_by(
        db_session, field="key", value=DEFAULT_ACCESS_POLICY
    )
    if not access_policy:
        log.info(f"Creating default policy: {DEFAULT_ACCESS_POLICY}...")
        access_policy = Policy.create(
            db=db_session,
            data={
                "name": "Default Access Policy",
                "key": DEFAULT_ACCESS_POLICY,
                "execution_timeframe": 45,
                "client_id": client_id,
                "drp_action": DrpAction.access.value,
            },
        )
    else:
        log.debug(
            f"Skipping {DEFAULT_ACCESS_POLICY} creation as it already exists in the database"
        )

    access_rule: Optional[FidesBase] = Rule.get_by(
        db_session, field="key", value=DEFAULT_ACCESS_POLICY_RULE
    )
    if not access_rule:
        log.info(f"Creating default policy rule: {DEFAULT_ACCESS_POLICY_RULE}...")
        access_rule = Rule.create(
            db=db_session,
            data={
                "action_type": ActionType.access.value,
                "name": "Default Access Rule",
                "key": DEFAULT_ACCESS_POLICY_RULE,
                "policy_id": access_policy.id,
                "client_id": client_id,
            },
        )

        log.info("Creating default policy rules targets...")
        for target in default_data_categories:
            data = {
                "data_category": target,
                "rule_id": access_rule.id,
            }
            compound_key = to_snake_case(RuleTarget.get_compound_key(data=data))
            data["key"] = compound_key
            try:
                RuleTarget.create(
                    db=db_session,
                    data=data,
                )
            except KeyOrNameAlreadyExists:  # pragma: no cover
                # This rule target already exists against the Policy
                pass
    else:
        log.debug(
            f"Skipping {DEFAULT_ACCESS_POLICY_RULE} creation as it already exists in the database"
        )


def load_default_erasure_policy(
    db_session: Session, client_id: str, default_data_categories: List[str]
) -> None:
    erasure_policy: Optional[FidesBase] = Policy.get_by(
        db_session, field="key", value=DEFAULT_ERASURE_POLICY
    )
    if not erasure_policy:
        log.info(f"Creating default policy: {DEFAULT_ERASURE_POLICY}...")
        erasure_policy = Policy.create(
            db=db_session,
            data={
                "name": "Default Erasure Policy",
                "key": DEFAULT_ERASURE_POLICY,
                "execution_timeframe": 45,
                "client_id": client_id,
                "drp_action": DrpAction.deletion.value,
            },
        )
    else:
        log.debug(
            f"Skipping {DEFAULT_ERASURE_POLICY} creation as it already exists in the database"
        )

    erasure_rule: Optional[FidesBase] = Rule.get_by(
        db_session, field="key", value=DEFAULT_ERASURE_POLICY_RULE
    )
    if not erasure_rule:
        log.info(f"Creating default policy rule: {DEFAULT_ERASURE_POLICY_RULE}...")
        erasure_rule = Rule.create(
            db=db_session,
            data={
                "action_type": ActionType.erasure.value,
                "name": "Default Erasure Rule",
                "key": DEFAULT_ERASURE_POLICY_RULE,
                "policy_id": erasure_policy.id,
                "client_id": client_id,
                "masking_strategy": {
                    "strategy": DEFAULT_ERASURE_MASKING_STRATEGY,
                    "configuration": {},
                },
            },
        )

        log.info("Creating default policy rule targets...")
        for target in default_data_categories:
            data = {
                "data_category": target,
                "rule_id": erasure_rule.id,
            }
            compound_key = to_snake_case(RuleTarget.get_compound_key(data=data))
            data["key"] = compound_key
            try:
                RuleTarget.create(
                    db=db_session,
                    data=data,
                )
            except KeyOrNameAlreadyExists:  # pragma: no cover
                # This rule target already exists against the Policy
                pass
    else:
        log.debug(
            f"Skipping {DEFAULT_ERASURE_POLICY_RULE} creation as it already exists in the database"
        )

    log.info(f"Creating default policy: {DEFAULT_CONSENT_POLICY}...")
    consent_policy = Policy.create_or_update(
        db=db_session,
        data={
            "name": "Default Consent Policy",
            "key": DEFAULT_CONSENT_POLICY,
            "execution_timeframe": 45,
            "client_id": client_id,
        },
    )

    log.info(f"Creating default policy rule: {DEFAULT_CONSENT_RULE}...")
    Rule.create_or_update(
        db=db_session,
        data={
            "action_type": ActionType.consent.value,
            "name": "Default Consent Rule",
            "key": DEFAULT_CONSENT_RULE,
            "policy_id": consent_policy.id,
            "client_id": client_id,
        },
    )


def load_default_dsr_policies(session: Session) -> None:
    """
    Checks whether DSR execution policies exist in the database, and
    inserts them to target a default set of data categories if not.
    """
    client_id = get_client_id(session)

    # By default, include all categories *except* those related to a user's
    # financial, payment, and credentials data. These are typically not
    # included in access and erasure requests as they are covered by other
    # compliance programs (e.g. legal, tax, security) and most
    # organizations need to be extra careful about how these are used -
    # especially for erasure! Therefore, a safe default for "out of the
    # box" behaviour is to exclude these

    default_data_categories = get_user_data_categories()

    log.debug(
        f"Preparing to create default rules for the following Data Categories: {default_data_categories} if they do not already exist"
    )

    load_default_access_policy(session, client_id, default_data_categories)
    load_default_erasure_policy(session, client_id, default_data_categories)

    log.info("All default policies & rules created")


def load_default_resources(session: Session) -> None:
    """
    Seed the database with default resources that the application
    expects to be available.
    """
    load_default_organization(session)
    load_default_taxonomy(session)
    load_default_dsr_policies(session)


async def load_samples(async_session: AsyncSession) -> None:
    # pylint: disable=too-many-branches, too-many-statements
    log.info("Loading sample resources...")
    try:
        sample_resources = dict(load_sample_resources_from_project())
        for resource_type, resources in sample_resources.items():
            if len(resources):
                if resource_type == "system":
                    await upsert_system(resources, async_session)
                else:
                    await upsert_resources(
                        sql_model_map[resource_type],
                        [e.model_dump(mode="json") for e in resources],
                        async_session,
                    )
    except QueryError:  # pragma: no cover
        pass  # The upsert_resources function will log any error

    log.info("Loading sample connections...")
    try:
        sample_connections = load_sample_connections_from_project()
        with sync_session() as db_session:
            for connection in sample_connections:
                assert connection.key, "Connection Key expected!"
                # If the connection config already exists, skip creation!
                # NOTE: This creates an edge case where the sample data was
                # created previously, but has since changed. By not deleting &
                # recreating here, we allow the "old" data to persist. That's an
                # acceptable risk here, so we log an INFO message to provide a
                # breadcrumb back to this code.
                connection_config: Optional[ConnectionConfig] = ConnectionConfig.get_by(
                    db=db_session, field="key", value=connection.key
                )
                if connection_config:
                    log.debug(
                        f"Sample connection '{connection.key}' already exists, skipping..."
                    )
                    continue

                # For SaaS connections, reuse our "instantiate from template" API
                if (
                    connection.connection_type == "saas"
                    and connection.saas_connector_type
                ):
                    log.debug(
                        f"Loading sample connection with SaaS connection type '{connection.saas_connector_type}'..."
                    )
                    saas_template_data = dict(connection)
                    saas_template_data["instance_key"] = connection.key
                    saas_template_data.pop(
                        "dataset", None
                    )  # not supported by this API!
                    saas_template_data.pop(
                        "system_key", None
                    )  # not supported by this API!
                    instantiate_connection_from_template(
                        db=db_session,
                        saas_connector_type=connection.saas_connector_type,
                        template_values=SaasConnectionTemplateValues.model_validate(
                            saas_template_data
                        ),
                    )

                    # Check that it succeeded!
                    connection_config = ConnectionConfig.get_by(
                        db=db_session, field="key", value=connection.key
                    )
                    if not connection_config:
                        log.debug(
                            f"Failed to create sample connection '{connection.key}'"
                        )
                    continue

                # For non-SaaS connections, reuse our connection & dataset APIs
                log.debug(
                    f"Loading sample connection with type '{connection.connection_type}'..."
                )
                connection_config_data = dict(connection)
                connection_config_data.pop(
                    "dataset", None
                )  # not supported by this API!
                connection_config_data.pop(
                    "system_key", None
                )  # not supported by this API!
                patch_connection_configs(
                    db=db_session,
                    configs=[
                        CreateConnectionConfigurationWithSecrets.model_validate(
                            connection_config_data
                        )
                    ],
                )

                # Check that it succeeded!
                connection_config = ConnectionConfig.get_by(
                    db=db_session, field="key", value=connection.key
                )
                if not connection_config:
                    log.debug(f"Failed to create sample connection '{connection.key}'")
                    continue

                # Create the DatasetConfig by linking to an existing Dataset
                dataset_key = connection.dataset
                if dataset_key:
                    log.debug(
                        f"Linking sample connection with key '{connection.key}' to dataset '{dataset_key}'..."
                    )
                    dataset = Dataset.get_by(
                        db=db_session, field="fides_key", value=dataset_key
                    )
                    if not dataset:
                        log.debug(
                            f"Could not find existing dataset '{dataset_key}' for sample connection '{connection.key}'"
                        )
                        continue

                    dataset_pair = DatasetConfigCtlDataset(
                        fides_key=dataset_key, ctl_dataset_fides_key=dataset_key
                    )
                    dataset_config_service = DatasetConfigService(db=db_session)
                    patch_dataset_configs(
                        dataset_pairs=[dataset_pair],
                        dataset_config_service=dataset_config_service,
                        connection_config=connection_config,
                    )
                    dataset_config = DatasetConfig.get_by(
                        db=db_session, field="fides_key", value=dataset_key
                    )
                    if not dataset_config:
                        log.debug(
                            f"Failed to create dataset config '{dataset_key}' for sample connection '{connection.key}'"
                        )
                        continue

                # Link the connection to an existing system
                system_key = connection.system_key
                if system_key:
                    log.debug(
                        f"Linking sample connection with key '{connection.key}' to system '{system_key}'..."
                    )
                    system = System.get_by(
                        db=db_session, field="fides_key", value=system_key
                    )
                    if not system:
                        log.debug(
                            f"Could not find existing system '{system_key}' for sample connection '{connection.key}'"
                        )
                        continue

                    linked_connection_config = ConnectionConfig.create_or_update(
                        db=db_session,
                        data={
                            **connection_config.__dict__,
                            **{"system_id": system.id},
                        },
                        check_name=False,
                    )
                    if not linked_connection_config:
                        log.debug(
                            f"Failed to link sample connection '{connection.key}' to system '{system_key}'"
                        )
                        continue

    except QueryError:  # pragma: no cover
        pass  # The upsert_resources function will log any error

    if os.getenv("MAILGUN_DOMAIN") and os.getenv("MAILGUN_API_KEY"):
        log.info("Loading Mailgun messaging config")
        with sync_session() as db:
            messaging_config = create_or_update_messaging_config(
                db=db,
                config=MessagingConfigRequest(
                    key="my_mailgun_config",
                    name="Mailgun",
                    service_type=MessagingServiceType.mailgun,
                    details=MessagingServiceDetailsMailgun(
                        domain=os.getenv("MAILGUN_DOMAIN")
                    ),
                ),
            )
            messaging_config.set_secrets(
                db=db,
                messaging_secrets=MessagingSecretsMailgunDocs(
                    mailgun_api_key=os.getenv("MAILGUN_API_KEY")
                ),
            )
            CONFIG.notifications.notification_service_type = "mailgun"
            ApplicationConfig.update_config_set(db, CONFIG)


def load_default_organization(db: Session) -> None:
    """
    Seed the database with a default organization unless
    one with a matching name already exists.
    """
    log.info("Loading the default organization...")
    organizations: List[Dict] = list(
        map(dict, DEFAULT_TAXONOMY.model_dump(mode="json")["organization"])
    )

    inserted = 0
    for org in organizations:
        try:
            existing = Organization.get_by(
                db=db, field="fides_key", value=org["fides_key"]
            )
            if not existing:
                Organization.create(db=db, data=org, check_name=False)
                inserted += 1
        except AlreadyExistsError:
            pass

    log.debug(f"INSERTED {inserted} organization resource(s)")
    log.debug(f"SKIPPED {len(organizations)-inserted} organization resource(s)")


def load_default_taxonomy(db: Session) -> None:
    """Synchronous version: Seed the database with the default taxonomy resources."""
    upsert_resource_types = list(DEFAULT_TAXONOMY.model_fields_set)
    upsert_resource_types.remove("organization")

    log.info("Loading the default fideslang taxonomy resources...")
    for resource_type in upsert_resource_types:
        sql_model = sql_model_map[resource_type]
        log.debug(f"Processing {resource_type} resources...")
        default_resources = DEFAULT_TAXONOMY.model_dump(mode="json")[resource_type]

        try:
            # Fetch existing keys first
            existing_keys = {item[0] for item in db.query(sql_model.fides_key).all()}
        except Exception as exc:
            log.error(f"Error fetching existing keys for {resource_type}: {exc}")
            continue  # Skip this resource type if we can't get existing keys

        # Filter for resources that don't exist
        resources_to_create = [
            resource
            for resource in default_resources
            if resource["fides_key"] not in existing_keys
        ]

        if not resources_to_create:
            log.debug(f"No new {resource_type} resources to add from default taxonomy.")
            continue

        created_count = 0
        for resource_data in resources_to_create:
            try:
                sql_model.create_or_update(db=db, data=resource_data)
                created_count += 1
            except Exception:
                pass

        if created_count > 0:
            log.debug(f"UPSERTED {created_count} {resource_type} resource(s)")



from typing import List, Optional

from pydantic import BaseModel, ConfigDict

from fides.api.schemas.enums.connection_category import ConnectionCategory
from fides.api.schemas.enums.integration_feature import IntegrationFeature


class SaaSDisplayInfo(BaseModel):
    """
    Optional display information for SAAS integrations to enhance frontend presentation.
    When not provided, smart defaults will be inferred based on the integration type.
    """

    category: Optional[ConnectionCategory] = None
    tags: Optional[List[str]] = None
    enabled_features: Optional[List[IntegrationFeature]] = None

    model_config = ConfigDict(use_enum_values=True)



"""Module that adds interactions with okta"""

import ast
from functools import update_wrapper
from typing import Any, Callable, List, Optional

from fideslang.models import System, SystemMetadata
from okta.client import Client as OktaClient
from okta.exceptions import OktaAPIException
from okta.models import Application as OktaApplication

from fides.connectors.models import (
    ConnectorAuthFailureException,
    ConnectorFailureException,
    OktaConfig,
)


def get_okta_client(okta_config: Optional[OktaConfig]) -> OktaClient:
    """
    Returns an Okta client for the given okta config. Authentication can
    also be handled through environment variables that the okta python sdk support.

    Enabled raiseException config to facilitate exception handling
    """
    config_dict = okta_config.model_dump(mode="json") if okta_config else {}
    config_dict["raiseException"] = True
    okta_client = OktaClient(config_dict)
    return okta_client


def handle_common_okta_errors(func: Callable) -> Callable:
    """
    Function decorator which handles common errors for okta calls.
    Classifies exceptions based on error codes returned by the client.
    For a full list of error codes see https://developer.okta.com/docs/reference/error-codes/
    """

    async def wrapper_func(*args, **kwargs) -> Any:  # type: ignore
        try:
            return await func(*args, **kwargs)
        except OktaAPIException as error:
            error_json = ast.literal_eval(str(error))
            # E0000004 - Authentication exception
            # E0000011 - Invalid token exception
            if error_json["errorCode"] in ["E0000004", "E0000011"]:
                raise ConnectorAuthFailureException(error_json["errorSummary"])
            raise ConnectorFailureException(error_json["errorSummary"])

    return update_wrapper(wrapper_func, func)


@handle_common_okta_errors
async def validate_credentials(okta_config: Optional[OktaConfig]) -> None:
    """
    Calls the list_applications api with a page limit of 1 to validate
    okta credentials.
    """
    client = get_okta_client(okta_config=okta_config)
    query_parameters = {"limit": "1"}
    await client.list_applications(query_parameters)


@handle_common_okta_errors
async def list_okta_applications(okta_client: OktaClient) -> List[OktaApplication]:
    """
    Returns a list of Okta applications. Iterates through each page returned by
    the client.
    """
    applications = []
    current_applications, resp, _ = await okta_client.list_applications()
    while True:
        applications.extend(current_applications)
        if resp.has_next():
            current_applications, _ = await resp.next()
        else:
            break
    return applications


def create_okta_systems(
    okta_applications: List[OktaApplication], organization_key: str
) -> List[System]:
    """
    Returns a list of system objects given a list of Okta applications
    """
    systems = [
        System(
            fides_key=application.id,
            name=application.name,
            fidesctl_meta=SystemMetadata(
                resource_id=application.id,
            ),
            description=f"Fides Generated Description for Okta Application: {application.label}",
            system_type="okta_application",
            organization_fides_key=organization_key,
            privacy_declarations=[],
        )
        for application in okta_applications
        if application.status and application.status == "ACTIVE"
    ]
    return systems



"""add audit_log_resource table

Revision ID: 587c53fe3e99
Revises: 8a71872089e4
Create Date: 2023-05-18 19:21:33.496435

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "587c53fe3e99"
down_revision = "8a71872089e4"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "audit_log_resource",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("user_id", sa.String(), nullable=True),
        sa.Column("request_path", sa.String(), nullable=True),
        sa.Column("request_type", sa.String(), nullable=True),
        sa.Column("fides_keys", sa.ARRAY(sa.String()), nullable=True),
        sa.Column("extra_data", sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_audit_log_resource_id"), "audit_log_resource", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_audit_log_resource_user_id"),
        "audit_log_resource",
        ["user_id"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        op.f("ix_audit_log_resource_user_id"), table_name="audit_log_resource"
    )
    op.drop_index(op.f("ix_audit_log_resource_id"), table_name="audit_log_resource")
    op.drop_table("audit_log_resource")
    # ### end Alembic commands ###



from __future__ import annotations

from enum import Enum
from typing import List, Optional

from sqlalchemy import ARRAY, Column
from sqlalchemy import Enum as SQLAlchemyEnum
from sqlalchemy import ForeignKey, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import Session, relationship

from fides.api.db.base_class import Base, FidesBase  # type: ignore[attr-defined]
from fides.api.models.detection_discovery.core import MonitorConfig
from fides.api.models.worker_task import (
    ExecutionLogStatus,
    TaskExecutionLog,
    WorkerTask,
)


class MonitorTaskType(Enum):
    """
    Types of tasks that can be executed by a worker.
    """

    DETECTION = "detection"
    CLASSIFICATION = "classification"
    PROMOTION = "promotion"
    REMOVAL_PROMOTION = "removal_promotion"


class MonitorTask(WorkerTask, Base):
    """
    A monitor task executed by a worker.
    """

    # celery_id is used to track task executions. While MonitorTask.id remains constant,
    # celery_id changes with each execution or retry of the task, allowing us to track
    # the current execution state while maintaining a stable reference to the original task.
    celery_id = Column(
        String(255), unique=True, nullable=False, default=FidesBase.generate_uuid
    )
    task_arguments = Column(JSONB, nullable=True)  # To be able to rerun the task
    # Contains info, warning, or error messages
    message = Column(String)
    monitor_config_id = Column(
        String,
        ForeignKey(MonitorConfig.id_field_path, ondelete="CASCADE"),
        index=True,
        nullable=False,
    )
    staged_resource_urns = Column(ARRAY(String), nullable=True)
    child_resource_urns = Column(ARRAY(String), nullable=True)

    monitor_config = relationship(MonitorConfig, cascade="all, delete")
    execution_logs = relationship(
        "MonitorTaskExecutionLog", back_populates="monitor_task", cascade="all, delete"
    )

    @classmethod
    def allowed_action_types(cls) -> List[str]:
        return [e.value for e in MonitorTaskType]


class TaskRunType(Enum):
    """
    Type of task run.
    """

    MANUAL = "manual"
    SYSTEM = "system"


class MonitorTaskExecutionLog(TaskExecutionLog, Base):
    """
    Stores the individual execution logs associated with a MonitorTask.
    """

    # This celery_id preserves the specific execution ID for historical tracking,
    # unlike MonitorTask.celery_id which is updated with each execution.
    # This allows us to maintain a complete history of all task execution attempts.
    celery_id = Column(String(255), nullable=False)
    monitor_task_id = Column(
        String,
        ForeignKey(MonitorTask.id_field_path, ondelete="CASCADE"),
        index=True,
        nullable=False,
    )
    run_type = Column(
        SQLAlchemyEnum(TaskRunType), nullable=False, default=TaskRunType.SYSTEM
    )

    monitor_task = relationship("MonitorTask", back_populates="execution_logs")


def create_monitor_task_with_execution_log(
    db: Session, monitor_task_data: dict
) -> MonitorTask:
    """
    Creates a monitor task with an execution log.
    The default status is pending for the task and pending for the execution log.
    """
    status = ExecutionLogStatus.pending
    task_record = MonitorTask(  # type: ignore
        status=status.value,
        **monitor_task_data,
    )
    db.add(task_record)
    db.flush()

    execution_log = MonitorTaskExecutionLog(  # type: ignore
        monitor_task=task_record, celery_id=task_record.celery_id, status=status
    )
    db.add(execution_log)

    db.commit()
    db.refresh(task_record)
    return task_record


def update_monitor_task_with_execution_log(
    db: Session,
    status: ExecutionLogStatus,
    task_record: Optional[MonitorTask] = None,
    celery_id: Optional[str] = None,
    message: Optional[str] = None,
    run_type: TaskRunType = TaskRunType.SYSTEM,
) -> MonitorTask:
    """
    Updates a monitor task with an execution log.

    It must be either celery_id or task_record. If it doesn't receive a celery_id, it's assumed a new one needs to be created because a new run is about to be performed.
    If it receives a celery_id, it means it only needs to update the status of an existing run. It can receive task_record to avoid querying the database again to get it.
    """
    if not celery_id and not task_record:
        raise ValueError("Either celery_id or task_record must be provided")

    if celery_id and not task_record:
        task_record = MonitorTask.get_by(db=db, field="celery_id", value=celery_id)
        if not task_record:
            raise ValueError(f"Could not find MonitorTask with celery_id {celery_id}")

    assert task_record is not None  # help type checker understand the control flow

    if not celery_id:
        celery_id = task_record.generate_uuid()
        task_record.celery_id = celery_id

    task_record.status = status.value  # type: ignore
    task_record.message = message

    MonitorTaskExecutionLog(  # type: ignore
        monitor_task=task_record,
        status=status,
        message=message,
        celery_id=celery_id,
        run_type=run_type,
    )

    db.commit()
    db.refresh(task_record)
    return task_record



"""This module handles finding and parsing fides configuration files."""

# pylint: disable=C0115,C0116, E0213
from typing import Dict

from pydantic import Field
from pydantic_settings import SettingsConfigDict

from fides.core.utils import create_auth_header, get_auth_header

from .fides_settings import FidesSettings

ENV_PREFIX = "FIDES__USER__"


def try_get_auth_header() -> Dict[str, str]:
    """Try to get the auth header. If an error is thrown, return a default auth header instead."""
    try:
        return get_auth_header(verbose=False)
    except SystemExit:
        return create_auth_header("defaulttoken")


class UserSettings(FidesSettings):
    """Configuration settings that apply to the current user as opposed to the entire application instance."""

    auth_header: Dict[str, str] = Field(
        default=try_get_auth_header(),
        description="Authentication header built automatically from the credentials file.",
        exclude=True,
    )
    analytics_opt_out: bool = Field(
        default=True,
        description="When set to true, prevents sending privacy-respecting anonymous analytics data to Ethyca.",
    )
    encryption_key: str = Field(
        default="test_encryption_key",
        description="An arbitrary string used to encrypt the user data stored in the database. Encryption is implemented using PGP.",
    )
    username: str = Field(
        default="", description="The username used to log into the Fides webserver."
    )
    password: str = Field(
        default="", description="The password used to log into the Fides webserver."
    )
    model_config = SettingsConfigDict(env_prefix=ENV_PREFIX)



"""EventAudit model and related enums for comprehensive audit logging."""

from enum import Enum as EnumType

from sqlalchemy import Column, String, Text
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declared_attr

from fides.api.db.base_class import Base
from fides.api.db.util import EnumColumn


class EventAuditType(str, EnumType):
    """Hierarchical event types for audit logging - variable depth as needed."""

    # System
    system_updated = "system.updated"

    # System group
    system_group_created = "system_group.created"
    system_group_updated = "system_group.updated"
    system_group_deleted = "system_group.deleted"

    # Taxonomy
    taxonomy_created = "taxonomy.created"
    taxonomy_updated = "taxonomy.updated"
    taxonomy_deleted = "taxonomy.deleted"
    taxonomy_usage_updated = "taxonomy.usage.updated"
    taxonomy_usage_deleted = "taxonomy.usage.deleted"
    taxonomy_element_created = "taxonomy.element.created"
    taxonomy_element_updated = "taxonomy.element.updated"
    taxonomy_element_deleted = "taxonomy.element.deleted"


class EventAuditStatus(str, EnumType):
    """Status enum for event audit logging."""

    succeeded = "succeeded"
    failed = "failed"


class EventAudit(Base):
    """Audit log for significant business events across the Fides platform."""

    @declared_attr
    def __tablename__(cls) -> str:
        return "event_audit"

    # Uses EventAuditType values but left as String to avoid future migrations
    event_type = Column(String, index=True, nullable=False)
    user_id = Column(String, nullable=True, index=True)

    # Resource information
    resource_type = Column(String, nullable=True, index=True)
    resource_identifier = Column(String, nullable=True, index=True)

    # User-friendly description of event
    description = Column(Text, nullable=True)

    # Structured data about event
    event_details = Column(JSONB, nullable=True)

    # Status of the event
    status = Column(
        EnumColumn(EventAuditStatus),
        index=True,
        nullable=False,
    )



from typing import Callable, List, Optional, Tuple

from celery.app.task import Task
from loguru import logger
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import Query, Session
from tenacity import (
    RetryCallState,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from fides.api.common_exceptions import (
    PrivacyRequestCanceled,
    PrivacyRequestNotFound,
    RequestTaskNotFound,
    ResumeTaskException,
    UpstreamTasksNotReady,
)
from fides.api.graph.config import TERMINATOR_ADDRESS, CollectionAddress
from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.models.privacy_request import ExecutionLog, PrivacyRequest, RequestTask
from fides.api.models.worker_task import ExecutionLogStatus
from fides.api.schemas.policy import ActionType, CurrentStep
from fides.api.schemas.privacy_request import PrivacyRequestStatus
from fides.api.task.graph_task import (
    GraphTask,
    mark_current_and_downstream_nodes_as_failed,
)
from fides.api.task.manual.manual_task_address import ManualTaskAddress
from fides.api.task.manual.manual_task_graph_task import ManualTaskGraphTask
from fides.api.task.task_resources import TaskResources
from fides.api.tasks import DSR_QUEUE_NAME, DatabaseTask, celery_app
from fides.api.util.cache import cache_task_tracking_key
from fides.api.util.collection_util import Row
from fides.api.util.logger_context_utils import LoggerContextKeys, log_context
from fides.api.util.memory_watchdog import memory_limiter

# DSR 3.0 task functions


def get_privacy_request_and_task(
    session: Session, privacy_request_id: str, privacy_request_task_id: str
) -> Tuple[PrivacyRequest, RequestTask]:
    """
    Retrieves and validates a privacy request and its associated task
    """

    privacy_request: Optional[PrivacyRequest] = PrivacyRequest.get(
        db=session, object_id=privacy_request_id
    )
    request_task: Optional[RequestTask] = RequestTask.get(
        db=session, object_id=privacy_request_task_id
    )

    if not privacy_request:
        raise PrivacyRequestNotFound(
            f"Privacy request with id {privacy_request_id} not found"
        )

    if privacy_request.status == PrivacyRequestStatus.canceled:
        raise PrivacyRequestCanceled(
            f"Cannot execute request task {privacy_request_task_id} of privacy request {privacy_request_id}: status is {privacy_request.status.value}"
        )

    if not request_task or not request_task.privacy_request_id == privacy_request.id:
        raise RequestTaskNotFound(
            f"Request Task with id {privacy_request_task_id} not found for privacy request {privacy_request_id}"
        )

    return privacy_request, request_task


def run_prerequisite_task_checks(
    session: Session, privacy_request_id: str, privacy_request_task_id: str
) -> Tuple[PrivacyRequest, RequestTask, Query]:
    """
    Upfront checks that run as soon as the RequestTask is executed by the worker.

    Returns resources for use in executing a task
    """

    privacy_request, request_task = get_privacy_request_and_task(
        session, privacy_request_id, privacy_request_task_id
    )

    assert request_task  # For mypy

    upstream_results: Query = request_task.upstream_tasks_objects(session)

    # Only bother running this if the current task body needs to run
    if request_task.status == ExecutionLogStatus.pending:
        # Only running the upstream check instead of RequestTask.can_queue_request_task since
        # the node is already queued.
        if not request_task.upstream_tasks_complete(session, should_log=False):
            raise UpstreamTasksNotReady(
                f"Cannot start {request_task.action_type} task {request_task.collection_address}. Waiting for upstream tasks to finish."
            )

    return privacy_request, request_task, upstream_results


def create_graph_task(
    session: Session, request_task: RequestTask, resources: TaskResources
) -> GraphTask:
    """Hydrates a GraphTask from the saved collection details on the Request Task in the database

    This could fail if things like our Collection definitions have changed since we created the Task
    to begin with - this may be unrecoverable and a new Privacy Request should be created.
    """
    try:
        collection_address = request_task.request_task_address

        # Check if this is a manual task address
        graph_task: GraphTask
        if ManualTaskAddress.is_manual_task_address(collection_address):
            graph_task = ManualTaskGraphTask(resources)
        else:
            graph_task = GraphTask(resources)

    except Exception as exc:
        logger.debug(
            "Cannot execute task - error loading task from database. Exception {}",
            str(exc),
        )
        # Normally the GraphTask takes care of creating the ExecutionLog, but in this case we can't create it in the first place!
        ExecutionLog.create(
            db=session,
            data={
                "connection_key": None,
                "dataset_name": request_task.dataset_name,
                "collection_name": request_task.collection_name,
                "fields_affected": [],
                "action_type": request_task.action_type,
                "status": ExecutionLogStatus.error,
                "privacy_request_id": request_task.privacy_request_id,
                "message": str(exc),
            },
        )
        mark_current_and_downstream_nodes_as_failed(request_task, session)

        raise ResumeTaskException(
            f"Cannot resume request task. Error hydrating task from database: Request Task {request_task.id} for Privacy Request {request_task.privacy_request_id}. {exc}"
        )

    return graph_task


def can_run_task_body(
    request_task: RequestTask,
) -> bool:
    """Return True if we can execute the task body. We should skip if the task is already
    complete or this is a root/terminator node"""
    if request_task.is_terminator_task:
        logger.info(
            "Terminator {} task reached.",
            request_task.action_type,
        )
        return False
    if request_task.is_root_task:
        # Shouldn't be possible but adding as a catch-all
        return False
    if request_task.status != ExecutionLogStatus.pending:
        logger_method(request_task)(
            "Skipping {} task {} with status {}.",
            request_task.action_type,
            request_task.collection_address,
            request_task.status.value,
        )
        return False

    return True


def log_retry_attempt(retry_state: RetryCallState) -> None:
    """Log queue_downstream_tasks retry attempts."""

    logger.warning(
        "queue_downstream_tasks attempt {} failed. Retrying in {} seconds...",
        retry_state.attempt_number,
        retry_state.next_action.sleep,  # type: ignore[union-attr]
    )


@retry(
    stop=stop_after_attempt(5),
    wait=wait_exponential(multiplier=1, min=1),
    retry=retry_if_exception_type(OperationalError),
    before_sleep=log_retry_attempt,
)
def queue_downstream_tasks_with_retries(
    database_task: DatabaseTask,
    privacy_request_id: str,
    privacy_request_task_id: str,
    current_step: CurrentStep,
    privacy_request_proceed: bool,
) -> None:
    with database_task.get_new_session() as session:
        privacy_request, request_task = get_privacy_request_and_task(
            session, privacy_request_id, privacy_request_task_id
        )
        with logger.contextualize(
            privacy_request_source=(
                privacy_request.source.value if privacy_request.source else None
            )
        ):
            log_task_complete(request_task)
            queue_downstream_tasks(
                session,
                request_task,
                privacy_request,
                current_step,
                privacy_request_proceed,
            )


def queue_downstream_tasks(
    session: Session,
    request_task: RequestTask,
    privacy_request: PrivacyRequest,
    next_step: CurrentStep,
    privacy_request_proceed: bool,
) -> None:
    """Queue downstream tasks of the current node **if** the downstream task has all its upstream tasks completed.

    If we've reached the terminator task, restart the privacy request from the appropriate checkpoint.
    """
    pending_downstream: Query = request_task.get_pending_downstream_tasks(session)
    for downstream_task in pending_downstream:
        if downstream_task.can_queue_request_task(session, should_log=True):
            log_task_queued(downstream_task, request_task.collection_address)
            queue_request_task(downstream_task, privacy_request_proceed)

    if (
        request_task.request_task_address == TERMINATOR_ADDRESS
        and request_task.status != ExecutionLogStatus.complete
    ):
        # Only queue privacy request from the next step if we haven't reached the terminator before.
        # Multiple pathways could mark the same node as complete, so we may have already reached the
        # terminator node through a quicker path.
        from fides.service.privacy_request.privacy_request_service import (  # pylint: disable=cyclic-import
            queue_privacy_request,
        )

        if (
            privacy_request_proceed
        ):  # For Testing, this could be set to False, so we could just
            # run one of the graphs and not the entire privacy request
            queue_privacy_request(
                privacy_request_id=privacy_request.id,
                from_step=next_step.value,
            )
        request_task.update_status(session, ExecutionLogStatus.complete)


@celery_app.task(base=DatabaseTask, bind=True)
@memory_limiter
@log_context(
    capture_args={
        "privacy_request_id": LoggerContextKeys.privacy_request_id,
        "privacy_request_task_id": LoggerContextKeys.task_id,
    }
)
def run_access_node(
    self: DatabaseTask,
    privacy_request_id: str,
    privacy_request_task_id: str,
    privacy_request_proceed: bool = True,
) -> None:
    """Run an individual task in the access graph for DSR 3.0 and queue downstream nodes
    upon completion if applicable"""

    try:
        with self.get_new_session() as session:
            privacy_request, request_task, upstream_results = (
                run_prerequisite_task_checks(
                    session, privacy_request_id, privacy_request_task_id
                )
            )
            with logger.contextualize(
                privacy_request_source=(
                    privacy_request.source.value if privacy_request.source else None
                )
            ):
                log_task_starting(request_task)

                if can_run_task_body(request_task):
                    # Build GraphTask resource to facilitate execution
                    with TaskResources(
                        privacy_request,
                        privacy_request.policy,
                        session.query(ConnectionConfig).all(),
                        request_task,
                        session,
                    ) as resources:
                        graph_task: GraphTask = create_graph_task(
                            session, request_task, resources
                        )
                        # Currently, upstream tasks and "input keys" (which are built by data dependencies)
                        # are the same, but they may not be the same in the future.
                        upstream_access_data: List[List[Row]] = (
                            _build_upstream_access_data(
                                graph_task.execution_node.input_keys, upstream_results
                            )
                        )
                        # Run the main access function
                        graph_task.access_request(*upstream_access_data)

        queue_downstream_tasks_with_retries(
            self,
            privacy_request_id,
            privacy_request_task_id,
            CurrentStep.upload_access,
            privacy_request_proceed,
        )
    except Exception as e:
        logger.error(f"Error in run_access_node: {e}")
        raise


@celery_app.task(base=DatabaseTask, bind=True)
@memory_limiter
@log_context(
    capture_args={
        "privacy_request_id": LoggerContextKeys.privacy_request_id,
        "privacy_request_task_id": LoggerContextKeys.task_id,
    }
)
def run_erasure_node(
    self: DatabaseTask,
    privacy_request_id: str,
    privacy_request_task_id: str,
    privacy_request_proceed: bool = True,
) -> None:
    """Run an individual task in the erasure graph for DSR 3.0 and queue downstream nodes
    upon completion if applicable"""
    with self.get_new_session() as session:
        privacy_request, request_task, _ = run_prerequisite_task_checks(
            session, privacy_request_id, privacy_request_task_id
        )
        with logger.contextualize(
            privacy_request_source=(
                privacy_request.source.value if privacy_request.source else None
            )
        ):
            log_task_starting(request_task)

            if can_run_task_body(request_task):
                with TaskResources(
                    privacy_request,
                    privacy_request.policy,
                    session.query(ConnectionConfig).all(),
                    request_task,
                    session,
                ) as resources:
                    # Build GraphTask resource to facilitate execution
                    erasure_graph_task: GraphTask = create_graph_task(
                        session, request_task, resources
                    )
                    # Get access data that was saved in the erasure format that was collected from the
                    # access task for the same collection.  This data is used to build the masking request
                    retrieved_data: List[Row] = (
                        request_task.get_data_for_erasures() or []
                    )

                    upstream_access_data: List[List[Row]] = []

                    try:
                        upstream_access_data = (
                            get_upstream_access_data_for_erasure_task(
                                request_task, session, resources
                            )
                        )
                    except Exception as e:
                        logger.error(
                            f"Unable to get upstream access data for erasure task {request_task.collection_address}: {e}"
                        )

                    # Run the main erasure function, passing along the upstream access data.
                    # The extra data is currently only needed for generating BigQuery delete statements.
                    erasure_graph_task.erasure_request(
                        retrieved_data, inputs=upstream_access_data
                    )

    queue_downstream_tasks_with_retries(
        self,
        privacy_request_id,
        privacy_request_task_id,
        CurrentStep.finalize_erasure,
        privacy_request_proceed,
    )


@celery_app.task(base=DatabaseTask, bind=True)
@memory_limiter
@log_context(
    capture_args={
        "privacy_request_id": LoggerContextKeys.privacy_request_id,
        "privacy_request_task_id": LoggerContextKeys.task_id,
    }
)
def run_consent_node(
    self: DatabaseTask,
    privacy_request_id: str,
    privacy_request_task_id: str,
    privacy_request_proceed: bool = True,
) -> None:
    """Run an individual task in the consent graph for DSR 3.0 and queue downstream nodes
    upon completion if applicable"""
    with self.get_new_session() as session:
        privacy_request, request_task, upstream_results = run_prerequisite_task_checks(
            session, privacy_request_id, privacy_request_task_id
        )
        with logger.contextualize(
            privacy_request_source=(
                privacy_request.source.value if privacy_request.source else None
            )
        ):
            log_task_starting(request_task)

            if can_run_task_body(request_task):
                # Build GraphTask resource to facilitate execution
                with TaskResources(
                    privacy_request,
                    privacy_request.policy,
                    session.query(ConnectionConfig).all(),
                    request_task,
                    session,
                ) as resources:
                    graph_task: GraphTask = create_graph_task(
                        session, request_task, resources
                    )
                    access_data: List = []
                    if upstream_results:
                        # For consent, expected that there is only one upstream node, the root node,
                        # and it holds the identity data (stored in a list for consistency with other
                        # data stored in access_data)
                        access_data = upstream_results[0].get_access_data() or []

                    graph_task.consent_request(access_data[0] if access_data else {})

    queue_downstream_tasks_with_retries(
        self,
        privacy_request_id,
        privacy_request_task_id,
        CurrentStep.finalize_consent,
        privacy_request_proceed,
    )


def logger_method(request_task: RequestTask) -> Callable:
    """Log selected no-op items with debug method and others with info method"""
    return (
        logger.debug
        if request_task.status == ExecutionLogStatus.complete
        else logger.info
    )


def log_task_starting(request_task: RequestTask) -> None:
    """Convenience method for logging task start"""
    logger_method(request_task)(
        "Starting '{}' task {} with current status '{}'.",
        request_task.action_type,
        request_task.collection_address,
        request_task.status.value,
    )


def log_task_complete(request_task: RequestTask) -> None:
    """Convenience method for logging task completion"""
    logger.info(
        "{} task {} is {}.",
        request_task.action_type.capitalize(),
        request_task.collection_address,
        request_task.status.value,
    )


def _order_tasks_by_input_key(
    input_keys: List[CollectionAddress], upstream_tasks: Query
) -> List[Optional[RequestTask]]:
    """Order tasks by input key. If task doesn't exist, add None in its place

    Data being passed to GraphTask.access_request is expected to have the same order
    as input keys so we know which data belongs to which upstream collection
    """
    tasks: List[Optional[RequestTask]] = []
    for key in input_keys:
        task = next(
            (
                upstream
                for upstream in upstream_tasks
                if upstream.collection_address == key.value
            ),
            None,
        )
        tasks.append(task)
    return tasks


def get_upstream_access_data_for_erasure_task(
    erasure_request_task: RequestTask,
    session: Session,
    resources: TaskResources,
) -> List[List[Row]]:
    """
    Retrieves upstream access data for a given erasure request task.

    This function finds the corresponding access task for the erasure task,
    creates a GraphTask to extract input keys, and builds the upstream access data
    needed for erasure operations (particularly for BigQuery delete statements).

    Args:
        erasure_request_task: The erasure task that needs upstream access data
        session: Database session for querying
        resources: Task resources for creating GraphTask

    Returns:
        List[List[Row]]: Upstream access data ordered by input keys

    Raises:
        Exception: If the corresponding access task cannot be found
    """
    # Get the corresponding access task for the current erasure task
    access_request_task = (
        session.query(RequestTask)
        .filter(
            RequestTask.privacy_request_id == erasure_request_task.privacy_request_id,
            RequestTask.collection_address == erasure_request_task.collection_address,
            RequestTask.action_type == ActionType.access,
        )
        .first()
    )

    if not access_request_task:
        raise Exception(
            f"Unable to find access request task for erasure task {erasure_request_task.collection_address}"
        )

    # Convert the request task to a GraphTask to get the input_keys
    access_graph_task: GraphTask = create_graph_task(
        session, access_request_task, resources
    )

    # Build and return the upstream access data
    return _build_upstream_access_data(
        access_graph_task.execution_node.input_keys,
        access_request_task.upstream_tasks_objects(session),
    )


def _build_upstream_access_data(
    input_keys: List[CollectionAddress],
    upstream_tasks: Query,
) -> List[List[Row]]:
    """
    Helper function to build the access data for the current node.
    The access data is passed in the same order as the input keys.
    If we don't have access data for an upstream node, return an empty list.
    """

    ordered_upstream: List[Optional[RequestTask]] = _order_tasks_by_input_key(
        input_keys, upstream_tasks
    )
    return [task.get_access_data() if task else [] for task in ordered_upstream]


mapping = {
    ActionType.access.value: run_access_node,
    ActionType.erasure.value: run_erasure_node,
    ActionType.consent.value: run_consent_node,
}


def queue_request_task(
    request_task: RequestTask, privacy_request_proceed: bool = True
) -> None:
    """Queues the RequestTask in Celery and caches the Celery Task ID"""
    celery_task_fn: Task = mapping[request_task.action_type]
    celery_task = celery_task_fn.apply_async(
        queue=DSR_QUEUE_NAME,
        kwargs={
            "privacy_request_id": request_task.privacy_request_id,
            "privacy_request_task_id": request_task.id,
            "privacy_request_proceed": privacy_request_proceed,
        },
    )
    cache_task_tracking_key(request_task.id, celery_task.task_id)


def log_task_queued(request_task: RequestTask, location: str) -> None:
    """Helper for logging that tasks are queued"""
    logger_method(request_task)(
        "Queuing {} task {} from {}.",
        request_task.action_type,
        request_task.collection_address,
        location,
    )



from typing import Any, Dict, Optional, Set

from fideslang.validation import FidesKey
from loguru import logger
from sqlalchemy.orm import Session

from fides.api.common_exceptions import StorageUploadError
from fides.api.graph.graph import DataCategoryFieldMapping
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.models.storage import StorageConfig
from fides.api.schemas.storage.storage import (
    FileNaming,
    ResponseFormat,
    StorageDetails,
    StorageType,
)
from fides.api.service.storage.streaming.s3.streaming_s3 import upload_to_s3_streaming
from fides.api.tasks.storage import upload_to_gcs, upload_to_local, upload_to_s3


def upload(
    db: Session,
    privacy_request: PrivacyRequest,
    data: Dict,
    storage_key: FidesKey,
    data_category_field_mapping: Optional[DataCategoryFieldMapping] = None,
    data_use_map: Optional[Dict[str, Set[str]]] = None,
) -> str:
    """
    Retrieves storage configs and calls appropriate upload method
    :param db: SQLAlchemy Session
    :param request_id: Request id
    :param data: Dict of data to upload
    :param storage_key: Key representing where to upload data
    :return str representing location of upload (url or simply a description of where to find the data)
    """
    logger.debug("upload called with storage_key: {}", storage_key)

    config: Optional[StorageConfig] = StorageConfig.get_by(
        db=db, field="key", value=storage_key
    )

    if config is None:
        logger.warning("Storage type not found: {}", storage_key)
        raise StorageUploadError(f"Storage type not found: {storage_key}")

    logger.debug(
        "Retrieved storage config: key={}, type={}, has_secrets={}",
        config.key,
        config.type,
        config.secrets is not None,
    )

    if not config.secrets:
        logger.warning("Storage config has no secrets!")

    uploader: Any = _get_uploader_from_config_type(config.type)  # type: ignore
    logger.debug(
        "Using uploader: {}",
        uploader.__name__ if hasattr(uploader, "__name__") else type(uploader),
    )

    return uploader(db, config, data, privacy_request)


def get_extension(resp_format: ResponseFormat, enable_streaming: bool = False) -> str:
    """
    Determine file extension for various response formats.

    CSV's and HTML reports are zipped together before uploading to s3.
    """
    if enable_streaming:
        return "zip"

    if resp_format == ResponseFormat.csv:
        return "zip"

    if resp_format == ResponseFormat.json:
        return "json"

    if resp_format == ResponseFormat.html:
        return "zip"

    raise NotImplementedError(f"No extension defined for {resp_format}")


def _construct_file_key(
    request_id: str, config: StorageConfig, enable_streaming: bool = False
) -> str:
    """Constructs file key based on desired naming convention and request id, e.g. 23847234.json"""
    naming = config.details.get(
        StorageDetails.NAMING.value, FileNaming.request_id.value
    )
    if naming != FileNaming.request_id.value:
        raise ValueError(f"File naming of {naming} not supported")

    return f"{request_id}.{get_extension(config.format, enable_streaming)}"  # type: ignore


def _get_uploader_from_config_type(storage_type: StorageType) -> Any:
    """Determines which uploader method to use based on storage type"""
    return {
        StorageType.s3.value: _s3_uploader,
        StorageType.local.value: _local_uploader,
        StorageType.gcs.value: _gcs_uploader,
    }[storage_type.value]


def _s3_uploader(
    _: Session,
    config: StorageConfig,
    data: Dict,
    privacy_request: PrivacyRequest,
) -> str:
    """
    Constructs necessary info needed for s3 before calling upload.
    If `enable_streaming` is configured in the storage config, we use a streaming approach for better memory efficiency.
    Otherwise we fall back to the traditional upload method.
    """
    logger.debug(
        "_s3_uploader called with config: key={}, type={}, has_secrets={}",
        config.key,
        config.type,
        config.secrets is not None,
    )

    enable_streaming = config.details.get(StorageDetails.ENABLE_STREAMING.value, False)
    file_key: str = _construct_file_key(privacy_request.id, config, enable_streaming)

    bucket_name = config.details[StorageDetails.BUCKET.value]
    auth_method = config.details[StorageDetails.AUTH_METHOD.value]
    document = None

    if enable_streaming:
        file_key = f"{privacy_request.id}.zip"
        # Use streaming upload for better memory efficiency
        logger.debug("Using streaming S3 upload for {}", file_key)
        return upload_to_s3_streaming(
            config.secrets,  # type: ignore
            data,
            bucket_name,
            file_key,
            config.format.value,  # type: ignore
            privacy_request,
            document,
            auth_method,
        )

    file_key = _construct_file_key(privacy_request.id, config)

    # Fall back to traditional upload method
    logger.debug("Using traditional S3 upload for {}", file_key)
    return upload_to_s3(
        config.secrets,  # type: ignore
        data,
        bucket_name,
        file_key,
        config.format.value,  # type: ignore
        privacy_request,
        document,
        auth_method,
    )


def _gcs_uploader(
    _: Session,
    config: StorageConfig,
    data: Dict,
    privacy_request: PrivacyRequest,
) -> str:
    """Constructs necessary info needed for Google Cloud Storage before calling upload"""
    file_key: str = _construct_file_key(privacy_request.id, config)

    bucket_name = config.details[StorageDetails.BUCKET.value]
    auth_method = config.details[StorageDetails.AUTH_METHOD.value]

    return upload_to_gcs(
        config.secrets,
        data,
        bucket_name,
        file_key,
        config.format.value,  # type: ignore
        privacy_request,
        auth_method,
    )


def _local_uploader(
    _: Session,
    config: StorageConfig,
    data: Dict,
    privacy_request: PrivacyRequest,
) -> str:
    """Uploads data to local storage, used for quick-start/demo purposes"""
    file_key: str = _construct_file_key(privacy_request.id, config)
    return upload_to_local(data, file_key, privacy_request, config.format.value)  # type: ignore



from typing import Any, Dict, List

import pydash
from requests import Response

from fides.api.common_exceptions import PrivacyRequestError
from fides.api.schemas.saas.strategy_configuration import PollingAsyncDSRConfiguration
from fides.api.service.async_dsr.async_dsr_strategy import AsyncDSRStrategy
from fides.api.service.connectors.saas.authenticated_client import AuthenticatedClient
from fides.api.util.collection_util import Row
from fides.api.util.saas_util import map_param_values


class PollingAsyncDSRStrategy(AsyncDSRStrategy):
    """
    Strategy for polling async DSR requests.
    """

    name = "polling"
    configuration_model = PollingAsyncDSRConfiguration

    def __init__(self, configuration: PollingAsyncDSRConfiguration):
        self.status_request = configuration.status_request
        self.status_path = configuration.status_path
        self.status_completed_value = configuration.status_completed_value
        self.result_request = configuration.result_request
        self.result_path = configuration.result_path

    def get_status_request(
        self,
        client: AuthenticatedClient,
        secrets: Dict[str, Any],
        identity_data: Dict[str, Any],
    ) -> bool:
        """Executes the status requests, and move forward if its true"""
        param_values = secrets.copy()
        param_values.update(identity_data)
        prepared_status_request = map_param_values(
            "status", "polling request", self.status_request, param_values
        )

        response: Response = client.send(prepared_status_request)

        if response.ok:
            status_path_value = pydash.get(response.json(), self.status_path)
            # If the status path value is a boolean, return it
            if isinstance(status_path_value, bool):
                return status_path_value

            # if the status path is a string, check if its the expected string value
            if isinstance(status_path_value, str):
                if status_path_value == self.status_completed_value:
                    return True

                return False

            # if the status path is a list, check if the first element is the expected string value
            if isinstance(status_path_value, list):
                if status_path_value[0] == self.status_completed_value:
                    return True

                return False
            # And if we cant recognize the type, raise an error
            raise PrivacyRequestError(
                f"Status request returned an unexpected value: {status_path_value}"
            )

        raise PrivacyRequestError(
            f"Status request failed with status code {response.status_code}"
        )

    def get_result_request(
        self,
        client: AuthenticatedClient,
        secrets: Dict[str, Any],
        identity_data: Dict[str, Any],
    ) -> List[Row]:
        """Build request to get the result of the async DSR process"""
        param_values = secrets.copy()
        param_values.update(identity_data)
        prepared_result_request = map_param_values(
            "result",
            "polling request",
            self.result_request,
            param_values,  # type: ignore
        )
        response: Response = client.send(prepared_result_request)
        if response.ok:
            result = pydash.get(response.json(), self.result_path)
            return result

        raise PrivacyRequestError(
            f"Result request failed with status code {response.status_code}"
        )



# pylint: disable=R0401, C0302

from __future__ import annotations

from enum import Enum as EnumType
from typing import TYPE_CHECKING, List, Optional, Tuple

from loguru import logger
from sqlalchemy import Boolean, Column, ForeignKey, Integer, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.mutable import MutableDict, MutableList
from sqlalchemy.orm import Query, Session, relationship
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesGcmEngine,
    StringEncryptedType,
)

from fides.api.db.base_class import Base, JSONTypeOverride  # type: ignore[attr-defined]
from fides.api.db.util import EnumColumn
from fides.api.graph.config import (
    ROOT_COLLECTION_ADDRESS,
    TERMINATOR_ADDRESS,
    CollectionAddress,
)
from fides.api.models.field_types import EncryptedLargeDataDescriptor
from fides.api.models.privacy_request.execution_log import (
    COMPLETED_EXECUTION_LOG_STATUSES,
)
from fides.api.models.worker_task import ExecutionLogStatus, WorkerTask
from fides.api.schemas.base_class import FidesSchema
from fides.api.schemas.policy import ActionType
from fides.api.util.cache import (
    FidesopsRedis,
    celery_tasks_in_flight,
    get_async_task_tracking_cache_key,
    get_cache,
)
from fides.api.util.collection_util import Row
from fides.config import CONFIG

if TYPE_CHECKING:
    from fides.api.models.privacy_request.privacy_request import PrivacyRequest


class TraversalDetails(FidesSchema):
    """Schema to format saving pre-calculated traversal details on RequestTask.traversal_details"""

    dataset_connection_key: str
    incoming_edges: List[Tuple[str, str]]
    outgoing_edges: List[Tuple[str, str]]
    input_keys: List[str]
    skipped_nodes: Optional[List[Tuple[str, str]]] = None

    # TODO: remove this method once we support custom request fields in DSR graph.
    @classmethod
    def create_empty_traversal(cls, connection_key: str) -> TraversalDetails:
        """
        Creates an "empty" TraversalDetails object that only has the dataset connection key set.
        This is a bit of a hacky workaround needed to implement the Dynamic Erasure Emails feature,
        but we should no longer need it once the custom_request_fields are included in our graph
        traversal
        """
        return cls(
            dataset_connection_key=connection_key,
            incoming_edges=[],
            outgoing_edges=[],
            input_keys=[],
            skipped_nodes=[],
        )


class AsyncTaskType(EnumType):
    """
    Enum for asynchronous RequestTask types

    These tasks differ from our regular tasks in that they are not finished on the same process that they are created,
    but rather are put into a pause state and then triggered to completion by a different process

    manual: A task that is manually triggered by the user
    polling: A task that is triggered by a polling mechanism
    callback: A task that is triggered by a callback from an external system

    """

    manual = "manual"
    polling = "polling"
    callback = "callback"


# TODO: At some point we will refactor this model to store all task types in a common table that links to tables with specific task attributes.
class RequestTask(WorkerTask, Base):
    """
    An individual Task for a Privacy Request.

    When we execute a PrivacyRequest, we build a graph by combining the current datasets with the identity data
    and we save the nodes (collections) in the graph as Request Tasks.

    Currently, we build access, erasure, and consent Request Tasks.
    """

    privacy_request_id = Column(
        String,
        ForeignKey("privacyrequest.id", ondelete="SET NULL"),
        nullable=True,
        index=True,
    )

    # Identifiers of this request task
    collection_address = Column(
        String, nullable=False, index=True
    )  # Of the format dataset_name:collection_name for convenience
    dataset_name = Column(String, nullable=False, index=True)
    collection_name = Column(String, nullable=False, index=True)

    upstream_tasks = Column(
        MutableList.as_mutable(JSONB)
    )  # List of collection address strings
    downstream_tasks = Column(
        MutableList.as_mutable(JSONB)
    )  # List of collection address strings
    all_descendant_tasks = Column(
        MutableList.as_mutable(JSONB)
    )  # All tasks that can be reached by the current task.  This is useful when this task fails,
    # and we can mark every single one of these as failed.

    # Raw data retrieved from an access request is stored here.  This contains all of the
    # intermediate data we retrieved, needed for downstream tasks, but hasn't been filtered
    # by data category for the end user.
    _access_data = Column(  # An encrypted JSON String - saved as a list of Rows
        "access_data",
        StringEncryptedType(
            type_in=JSONTypeOverride,
            key=CONFIG.security.app_encryption_key,
            engine=AesGcmEngine,
            padding="pkcs5",
        ),
    )

    # This is the raw access data saved in erasure format (with placeholders preserved) to perform a masking request.
    # First saved on the access node, and then copied to the corresponding erasure node.
    _data_for_erasures = Column(  # An encrypted JSON String - saved as a list of rows
        "data_for_erasures",
        StringEncryptedType(
            type_in=JSONTypeOverride,
            key=CONFIG.security.app_encryption_key,
            engine=AesGcmEngine,
            padding="pkcs5",
        ),
    )

    # Use descriptors for automatic external storage handling
    access_data = EncryptedLargeDataDescriptor(
        field_name="access_data", empty_default=[]
    )

    data_for_erasures = EncryptedLargeDataDescriptor(
        field_name="data_for_erasures", empty_default=[]
    )

    # Written after an erasure is completed
    rows_masked = Column(Integer)
    # Written after a consent request is completed - not all consent
    # connectors will end up sending a request
    consent_sent = Column(Boolean)

    # For async tasks awaiting callback
    callback_succeeded = Column(Boolean)
    # to recognize Polling async task
    async_type = Column(
        EnumColumn(AsyncTaskType, native_enum=True),
        nullable=True,
    )

    # Stores a serialized collection that can be transformed back into a Collection to help
    # execute the current task
    collection = Column(MutableDict.as_mutable(JSONB))
    # Stores key details from traversal.traverse in the format of TraversalDetails
    traversal_details = Column(MutableDict.as_mutable(JSONB))

    privacy_request = relationship(
        "PrivacyRequest",
        back_populates="request_tasks",
        uselist=False,
    )

    @property
    def request_task_address(self) -> CollectionAddress:
        """Convert the collection_address into Collection Address format"""
        return CollectionAddress.from_string(self.collection_address)

    @property
    def is_root_task(self) -> bool:
        """Convenience helper for asserting whether the task is a root task"""
        return self.request_task_address == ROOT_COLLECTION_ADDRESS

    @property
    def is_terminator_task(self) -> bool:
        """Convenience helper for asserting whether the task is a terminator task"""
        return self.request_task_address == TERMINATOR_ADDRESS

    @classmethod
    def allowed_action_types(cls) -> List[str]:
        return [e.value for e in ActionType]

    def get_cached_task_id(self) -> Optional[str]:
        """Gets the cached celery task ID for this request task."""
        cache: FidesopsRedis = get_cache()
        task_id = cache.get(get_async_task_tracking_cache_key(self.id))
        return task_id

    def cleanup_external_storage(self) -> None:
        """Clean up all external storage files for this request task"""
        # Access the descriptor from the class to call cleanup
        RequestTask.access_data.cleanup(self)
        RequestTask.data_for_erasures.cleanup(self)

    def get_access_data(self) -> List[Row]:
        """Helper to retrieve access data or default to empty list"""
        return self.access_data or []

    def get_data_for_erasures(self) -> List[Row]:
        """Helper to retrieve erasure data needed to build masking requests or default to empty list"""
        return self.data_for_erasures or []

    def delete(self, db: Session) -> None:
        """Override delete to cleanup external storage first"""
        self.cleanup_external_storage()
        super().delete(db)

    def update_status(self, db: Session, status: ExecutionLogStatus) -> None:
        """Helper method to update a task's status"""
        self.status = status
        self.save(db)

    def get_tasks_with_same_action_type(
        self, db: Session, collection_address_str: str
    ) -> Query:
        """Fetch task on the same privacy request and action type as current by collection address"""
        return db.query(RequestTask).filter(
            RequestTask.privacy_request_id == self.privacy_request_id,
            RequestTask.action_type == self.action_type,
            RequestTask.collection_address == collection_address_str,
        )

    def get_pending_downstream_tasks(self, db: Session) -> Query:
        """Returns the immediate downstream task objects that are still pending"""
        return db.query(RequestTask).filter(
            RequestTask.privacy_request_id == self.privacy_request_id,
            RequestTask.action_type == self.action_type,
            RequestTask.collection_address.in_(self.downstream_tasks or []),
            RequestTask.status == ExecutionLogStatus.pending,
        )

    def can_queue_request_task(self, db: Session, should_log: bool = False) -> bool:
        """Returns True if upstream tasks are complete and the current Request Task
        is not running in another celery task.

        This check ignores its database status - that is checked elsewhere.
        """
        return self.upstream_tasks_complete(
            db, should_log
        ) and not self.request_task_running(should_log)

    def upstream_tasks_complete(self, db: Session, should_log: bool = False) -> bool:
        """Determines if all of the upstream tasks of the current task are complete"""
        upstream_tasks: Query = self.upstream_tasks_objects(db)
        tasks_complete: bool = all(
            upstream_task.status in COMPLETED_EXECUTION_LOG_STATUSES
            for upstream_task in upstream_tasks
        ) and upstream_tasks.count() == len(self.upstream_tasks or [])

        if not tasks_complete and should_log:
            logger.debug(
                "Upstream tasks incomplete for {} task {}.",
                self.action_type,
                self.collection_address,
            )

        return tasks_complete

    def upstream_tasks_objects(self, db: Session) -> Query:
        """Returns Request Task objects that are upstream of the current Request Task"""
        upstream_tasks: Query = db.query(RequestTask).filter(
            RequestTask.privacy_request_id == self.privacy_request_id,
            RequestTask.collection_address.in_(self.upstream_tasks or []),
            RequestTask.action_type == self.action_type,
        )
        return upstream_tasks

    def request_task_running(self, should_log: bool = False) -> bool:
        """Returns a rough measure if the Request Task is already running -
        not 100% accurate.

        This is further only applicable if you are running workers and
        CONFIG.execution.task_always_eager=False. This is just an extra check to reduce possible
        over-scheduling, but it is also okay if the same node runs multiple times.
        """
        celery_task_id: Optional[str] = self.get_cached_task_id()
        if not celery_task_id:
            return False

        if should_log:
            logger.debug(
                "Celery Task ID {} found for {} task {}.",
                celery_task_id,
                self.action_type,
                self.collection_address,
            )

        task_in_flight: bool = celery_tasks_in_flight([celery_task_id])

        if task_in_flight and should_log:
            logger.debug(
                "Celery Task {} already processing for {} task {}.",
                celery_task_id,
                self.action_type,
                self.collection_address,
            )

        return task_in_flight



"""Update ConsentStatus enum

Revision ID: 67d0e389b003
Revises: f36ce1bde293
Create Date: 2025-09-02 19:32:05.025508

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "67d0e389b003"
down_revision = "f36ce1bde293"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("ALTER TYPE consentstatus ADD VALUE 'pre_consent'")
    op.execute("ALTER TYPE consentstatus ADD VALUE 'cmp_error'")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    # Remove any records that have the new enum values before dropping them
    op.execute(
        "UPDATE asset SET consent_status = 'unknown' WHERE consent_status IN ('pre_consent', 'cmp_error')"
    )

    # Recreate the enum without the new values
    op.execute("ALTER TYPE consentstatus RENAME TO consentstatus_old")
    op.execute(
        "CREATE TYPE consentstatus AS ENUM ('with_consent', 'without_consent', 'exempt', 'unknown')"
    )

    # Drop default, change type, and set default back
    op.execute("ALTER TABLE asset ALTER COLUMN consent_status DROP DEFAULT")
    op.execute(
        "ALTER TABLE asset ALTER COLUMN consent_status TYPE consentstatus USING "
        "consent_status::text::consentstatus"
    )
    op.execute("ALTER TABLE asset ALTER COLUMN consent_status SET DEFAULT 'unknown'")
    op.execute("DROP TYPE consentstatus_old")
    # ### end Alembic commands ###



from datetime import datetime
from typing import Any, Dict, List, Optional, Sequence

from fideslang.models import PrivacyDeclaration, System
from fideslang.validation import FidesKey
from pydantic import ConfigDict, Field
from pydantic.main import BaseModel

from fides.api.schemas.connection_configuration.connection_config import (
    ConnectionConfigurationResponse,
)
from fides.api.schemas.user import UserResponse


class PrivacyDeclarationResponse(PrivacyDeclaration):
    """Extension of base pydantic model to include DB `id` field in the response"""

    id: str = Field(
        description="The database-assigned ID of the privacy declaration on the system. This is meant to be a read-only field, returned only in API responses"
    )


class BasicSystemResponse(System):
    """
    Extension of base pydantic model to include additional fields on the DB model that
    are relevant in API responses.

    This is still meant to be a "lightweight" model that does not reference relationships
    that may require additional querying beyond the `System` db table.
    """

    created_at: datetime


class SystemResponse(BasicSystemResponse):
    """Extension of base pydantic response model to include additional relationship fields that
    may require extra DB queries, like `privacy_declarations`, connection_config fields and cookies.

    This response model is generally useful for an API that returns a detailed view of _single_
    System record. Attempting to return bulk results with this model can lead to n+1 query issues.
    """

    privacy_declarations: Sequence[PrivacyDeclarationResponse] = Field(  # type: ignore[assignment]
        description=PrivacyDeclarationResponse.__doc__,
    )

    connection_configs: Optional[ConnectionConfigurationResponse] = Field(
        description=ConnectionConfigurationResponse.__doc__,
    )

    data_stewards: Optional[List[UserResponse]] = Field(
        description="System managers of the current system",
    )


class SystemHistoryResponse(BaseModel):
    """Response schema for a single system history entry"""

    edited_by: Optional[str] = None
    system_id: str
    before: Dict[str, Any]
    after: Dict[str, Any]
    created_at: datetime
    model_config = ConfigDict(from_attributes=True)


class AssignStewardRequest(BaseModel):
    """Request schema for assigning a data steward to one or more systems."""

    data_steward: str
    system_keys: List[FidesKey]



from typing import Any, Dict, List, Optional

from cassandra import AuthenticationFailed, UnresolvableContactPoints
from cassandra.auth import PlainTextAuthProvider
from cassandra.cluster import Cluster, NoHostAvailable, ResultSet, Session
from cassandra.protocol import SyntaxException
from cassandra.query import dict_factory
from loguru import logger

from fides.api.common_exceptions import ConnectionException
from fides.api.graph.execution import ExecutionNode
from fides.api.models.connectionconfig import ConnectionTestStatus
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest, RequestTask
from fides.api.schemas.connection_configuration.connection_secrets_scylla import (
    ScyllaSchema,
)
from fides.api.service.connectors.base_connector import BaseConnector
from fides.api.service.connectors.scylla_query_config import ScyllaDBQueryConfig
from fides.api.util.collection_util import Row
from fides.api.util.scylla_util import scylla_to_native_python


class ScyllaConnectorMissingKeyspace(Exception):
    pass


class ScyllaConnector(BaseConnector[Cluster]):
    """Scylla Connector"""

    @property
    def requires_primary_keys(self) -> bool:
        """ScyllaDB requires primary keys for erasures."""
        return True

    def build_uri(self) -> str:
        """
        Builds URI - Not yet implemented
        """
        return ""

    def get_config(self) -> ScyllaSchema:
        return ScyllaSchema(**self.configuration.secrets or {})

    def create_client(self) -> Cluster:
        """Returns a Scylla cluster"""

        config = self.get_config()

        auth_provider = PlainTextAuthProvider(
            username=config.username, password=config.password
        )
        try:
            cluster = Cluster(
                [config.host], port=config.port, auth_provider=auth_provider
            )
        except UnresolvableContactPoints:
            raise ConnectionException("No host available.")

        return cluster

    def query_config(self, node: ExecutionNode) -> ScyllaDBQueryConfig:
        return ScyllaDBQueryConfig(node)

    def set_schema(self, connection: Session) -> None:
        """Optionally override to set the schema for a given cluster that
        persists through the entire session"""
        config = self.get_config()
        if not config.keyspace:
            raise ScyllaConnectorMissingKeyspace(
                f"No keyspace provided in the ScyllaDB configuration for connector {self.configuration.key}"
            )
        logger.info("Setting ScyllaDB search_path before retrieving data")
        connection.set_keyspace(config.keyspace)

    @staticmethod
    def result_to_rows(results: ResultSet) -> List[Row]:
        """Convert Scylla DB results to a list of dictionaries"""
        rows = []
        for row in results:
            processed_row = {
                column_name: scylla_to_native_python(column_value)
                for column_name, column_value in row.items()
            }
            rows.append(processed_row)

        return rows

    def test_connection(self) -> Optional[ConnectionTestStatus]:
        """
        Connects to the scylla database and issues a trivial query
        """
        logger.info("Starting test connection to {}", self.configuration.key)
        config = self.get_config()
        cluster = self.client()

        try:
            with cluster.connect(
                config.keyspace if config.keyspace else None
            ) as client:
                client.execute("select now() from system.local")
        except NoHostAvailable as exc:
            if "Unable to connect to any servers using keyspace" in str(exc):
                raise ConnectionException("Unknown keyspace.")
            try:
                error = list(exc.errors.values())[0]
            except Exception:
                raise ConnectionException("No host available.")

            if isinstance(error, AuthenticationFailed):
                raise ConnectionException("Authentication failed.")

            raise ConnectionException("No host available.")

        except SyntaxException:
            raise ConnectionException("Syntax exception.")
        except Exception:
            raise ConnectionException("Connection Error connecting to Scylla DB.")

        return ConnectionTestStatus.succeeded

    def retrieve_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        input_data: Dict[str, List[Any]],
    ) -> List[Row]:
        """Retrieve ScyllaDB data"""
        query_config = self.query_config(node)
        client = self.client()
        generated_query = query_config.generate_query(input_data, policy)
        if generated_query is None:
            return []

        statement, params = generated_query
        logger.info("Starting data retrieval for {}", node.address)

        with client.connect() as connection:
            connection.row_factory = dict_factory
            self.set_schema(connection)
            results = connection.execute(statement, parameters=params)
            return self.result_to_rows(results)

    def mask_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        rows: List[Row],
        input_data: Optional[Dict[str, List[Any]]] = None,
    ) -> int:
        """Execute a masking request"""
        query_config = self.query_config(node)
        update_ct = 0
        client = self.client()
        for row in rows:
            generated_update_stmt = query_config.generate_update_stmt(
                row, policy, privacy_request
            )
            if generated_update_stmt is not None:
                statement, params = generated_update_stmt
                with client.connect() as connection:
                    self.set_schema(connection)
                    connection.execute(statement, parameters=params)
                    # Scylla doesn't return the number of rows updated by the executed statement
                    # so we just count the number of updates we've done, which should generally be the same
                    update_ct += 1
        return update_ct

    def close(self) -> None:
        """Close any held resources"""
        client = self.client()
        client.shutdown()



from typing import Any, Optional

from loguru import logger
from pydantic.v1.utils import deep_update

from fides.api.common_exceptions import AwaitingAsyncTask
from fides.api.models.attachment import AttachmentType
from fides.api.models.manual_task import (
    ManualTask,
    ManualTaskConfigurationType,
    ManualTaskEntityType,
    ManualTaskFieldType,
    ManualTaskInstance,
    ManualTaskSubmission,
    StatusType,
)
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.models.worker_task import ExecutionLogStatus
from fides.api.schemas.policy import ActionType
from fides.api.schemas.privacy_request import PrivacyRequestStatus
from fides.api.task.conditional_dependencies.logging_utils import (
    format_evaluation_failure_message,
    format_evaluation_success_message,
)
from fides.api.task.graph_task import GraphTask, retry
from fides.api.task.manual.manual_task_address import ManualTaskAddress
from fides.api.task.manual.manual_task_conditional_evaluation import (
    evaluate_conditional_dependencies,
    extract_conditional_dependency_data_from_inputs,
)
from fides.api.task.manual.manual_task_utils import (
    get_manual_task_for_connection_config,
)
from fides.api.task.task_resources import TaskResources
from fides.api.util.collection_util import Row


class ManualTaskGraphTask(GraphTask):
    """GraphTask implementation for ManualTask execution"""

    # class level constants
    DRY_RUN_PLACEHOLDER_VALUE = 1

    def __init__(self, resources: TaskResources) -> None:
        super().__init__(resources)
        self.connection_key = ManualTaskAddress.get_connection_key(
            self.execution_node.address
        )

    # ------------------------------------------------------------------------------------------------
    # Public methods
    # ------------------------------------------------------------------------------------------------

    def dry_run_task(self) -> int:
        """Return estimated row count for dry run - manual tasks don't have predictable counts"""
        return self.DRY_RUN_PLACEHOLDER_VALUE

    @retry(action_type=ActionType.access, default_return=[])
    def access_request(self, *inputs: list[Row]) -> list[Row]:
        """
        Execute manual task logic following the standard GraphTask pattern.
        Calls _run_request with ACCESS configs.
        Returns data if submitted, raise AwaitingAsyncTaskCallback if not
        """
        if self.resources.request.policy.get_action_type() == ActionType.erasure:
            # We're in an erasure privacy request's access phase - complete access task immediately
            # since access is just for data collection to support erasure, not for user data access
            self.update_status(
                "Access task completed immediately for erasure privacy request (data collection only)",
                [],
                ActionType.access,
                ExecutionLogStatus.complete,
            )
            return []

        result = self._run_request(
            ManualTaskConfigurationType.access_privacy_request,
            ActionType.access,
            *inputs,
        )
        if result is None:
            # Conditional skip or not applicable already logged upstream; do not mark complete here
            return []

        # We are picking up after awaiting input and have provided data – mark complete with record count
        self.log_end(ActionType.access, record_count=len(result))
        return result

    # Provide erasure support for manual tasks
    @retry(action_type=ActionType.erasure, default_return=0)
    def erasure_request(
        self,
        retrieved_data: list[Row],  # This is not used for manual tasks.
        *erasure_prereqs: int,  # noqa: D401, pylint: disable=unused-argument # TODO Remove when we stop support for DSR 2.0
        inputs: Optional[list[list[Row]]] = None,
    ) -> int:
        """Execute manual-task-driven erasure logic.
        Calls _run_request with ERASURE configs.

        Mirrors access_request behaviour but returns the number of rows masked (always 0)
        once all required manual task submissions are present. If submissions are
        incomplete the privacy request is paused awaiting user input.
        Returns the number of rows masked (always 0)
        Raises AwaitingAsyncTaskCallback if data is not submitted
        """
        if not inputs:
            inputs = []
        result = self._run_request(
            ManualTaskConfigurationType.erasure_privacy_request,
            ActionType.erasure,
            *inputs,
        )
        if result is None:
            # Conditional skip or not applicable already logged upstream; do not mark complete here
            return 0

        # Mark rows_masked = 0 (manual tasks do not mask data directly)
        if self.request_task.id:
            # Storing result for DSR 3.0; SQLAlchemy column typing triggers mypy warning
            self.request_task.rows_masked = 0  # type: ignore[assignment]

        # Picking up after awaiting input, mark erasure node complete with rows masked count (always 0)
        self.log_end(ActionType.erasure, record_count=0)
        return 0

    # ------------------------------------------------------------------------------------------------
    # Private methods
    # ------------------------------------------------------------------------------------------------

    def _run_request(
        self,
        config_type: ManualTaskConfigurationType,
        action_type: ActionType,
        *inputs: list[Row],
    ) -> Optional[list[Row]]:
        """
        Execute manual task logic following the standard GraphTask pattern:
        1. Create ManualTaskInstances if they don't exist
        2. Check if all required submissions are present
        3. Return data if submitted, raise AwaitingAsyncTask if not
        """
        manual_task = self._get_manual_task_or_none()
        if manual_task is None:
            return None

        # Complete a series of checks to determine if the manual task should be executed
        # If any of these checks fail, complete immediately or mark as skipped

        # Check if any eligible manual tasks have applicable configs
        if not self._check_manual_task_configs(manual_task, config_type, action_type):
            return None

        # Check if there are any rules for this action type
        if not self.resources.request.policy.get_rules_for_action(
            action_type=action_type
        ):
            return None

        # Extract conditional dependency data from inputs
        conditional_data = extract_conditional_dependency_data_from_inputs(
            *inputs, manual_task=manual_task, input_keys=self.execution_node.input_keys
        )
        # Evaluate conditional dependencies
        evaluation_result = evaluate_conditional_dependencies(
            self.resources.session, manual_task, conditional_data=conditional_data
        )
        detailed_message: Optional[str] = None
        # if there were conditional dependencies and they were not met,
        # clean up any existing ManualTaskInstances and return None to cause a skip
        if evaluation_result is not None and not evaluation_result.result:
            self._cleanup_manual_task_instances(manual_task, self.resources.request)
            detailed_message = format_evaluation_failure_message(evaluation_result)
            self.update_status(
                f"Manual task conditional dependencies not met. {detailed_message}",
                [],
                ActionType(self.resources.privacy_request_task.action_type),
                ExecutionLogStatus.skipped,
            )
            return None

        # Check/Create manual task instances for applicable configs only
        self._ensure_manual_task_instances(
            manual_task,
            self.resources.request,
            config_type,
        )

        # Check if all manual task instances have submissions for applicable configs only
        # No separate pending log; include details in the awaiting-processing log
        if evaluation_result:
            detailed_message = format_evaluation_success_message(evaluation_result)
        result = self._set_submitted_data_or_raise_awaiting_async_task_callback(
            manual_task,
            config_type,
            action_type,
            conditional_data=conditional_data,
            awaiting_detail_message=detailed_message,
        )
        return result

    def _check_manual_task_configs(
        self,
        manual_task: ManualTask,
        config_type: ManualTaskConfigurationType,
        action_type: ActionType,
    ) -> bool:
        has_access_configs = [
            config
            for config in manual_task.configs
            if config.is_current and config.config_type == config_type
        ]

        if not has_access_configs:
            # No access configs - complete immediately
            self.log_end(action_type)
            return False

        return True

    def _get_manual_task_or_none(self) -> Optional[ManualTask]:
        # Verify this is a manual task address
        if not ManualTaskAddress.is_manual_task_address(self.execution_node.address):
            raise ValueError(
                f"Invalid manual task address: {self.execution_node.address}"
            )

        # Get the manual task for this connection config (1:1 relationship)
        manual_task = get_manual_task_for_connection_config(
            self.resources.session, self.connection_key
        )
        return manual_task

    def _set_submitted_data_or_raise_awaiting_async_task_callback(
        self,
        manual_task: ManualTask,
        config_type: ManualTaskConfigurationType,
        action_type: ActionType,
        conditional_data: Optional[dict[str, Any]] = None,
        awaiting_detail_message: Optional[str] = None,
    ) -> Optional[list[Row]]:
        """
        Set submitted data for a manual task and raise AwaitingAsyncTaskCallback if all instances are not completed
        """
        # Check if all manual task instances have submissions for ACCESS configs only
        submitted_data = self._get_submitted_data(
            manual_task,
            self.resources.request,
            config_type,
            conditional_data=conditional_data,
        )

        if submitted_data is not None:
            result: list[Row] = [submitted_data] if submitted_data else []
            self.request_task.access_data = result

            return result

        # Set privacy request status to requires_input if not already set
        if self.resources.request.status != PrivacyRequestStatus.requires_input:
            self.resources.request.status = PrivacyRequestStatus.requires_input
            self.resources.request.save(self.resources.session)

        # This will trigger log_awaiting_processing via the @retry decorator; include conditional details
        base_msg = f"Manual task for {self.connection_key} requires user input"
        if awaiting_detail_message:
            base_msg = f"{base_msg}. {awaiting_detail_message}"
        raise AwaitingAsyncTask(base_msg)

    def _ensure_manual_task_instances(
        self,
        manual_task: ManualTask,
        privacy_request: PrivacyRequest,
        allowed_config_type: "ManualTaskConfigurationType",
    ) -> None:
        """Create ManualTaskInstances for configs matching `allowed_config_type` if they don't exist."""

        # ------------------------------------------------------------------
        # Check if instances already exist for this task & entity with the SAME config type
        # This prevents duplicates when configurations are versioned after the privacy
        # request has started, while allowing different config types (access vs erasure)
        # to have separate instances.
        # ------------------------------------------------------------------
        existing_task_instance = next(
            (
                instance
                for instance in privacy_request.manual_task_instances
                if instance.task_id == manual_task.id
                and instance.config.config_type == allowed_config_type
            ),
            None,
        )
        if existing_task_instance:
            # An instance already exists for this privacy request and config type – no need
            # to create another one tied to a newer config version.
            return

        # If no existing instances, create a new one for the current config
        # There will only be one config of each type per manual task
        # Sort by version descending to get the latest version first
        config = next(
            (
                config
                for config in sorted(
                    manual_task.configs,
                    key=lambda c: c.version if hasattr(c, "version") else 0,
                    reverse=True,
                )
                if config.is_current and config.config_type == allowed_config_type
            ),
            None,
        )

        if config:
            ManualTaskInstance.create(
                db=self.resources.session,
                data={
                    "task_id": manual_task.id,
                    "config_id": config.id,
                    "entity_id": privacy_request.id,
                    "entity_type": ManualTaskEntityType.privacy_request.value,
                    "status": StatusType.pending.value,
                },
            )

    def _get_submitted_data(
        self,
        manual_task: ManualTask,
        privacy_request: PrivacyRequest,
        allowed_config_type: "ManualTaskConfigurationType",
        conditional_data: Optional[dict[str, Any]] = None,
    ) -> Optional[dict[str, Any]]:
        """
        Check if all manual task instances have submissions for ALL fields and return aggregated data
        Returns None if any field submissions are missing (all fields must be completed or skipped)
        """
        candidate_instances: list[ManualTaskInstance] = [
            instance
            for instance in privacy_request.manual_task_instances
            if instance.task_id == manual_task.id
            and instance.config.config_type == allowed_config_type
        ]

        if not candidate_instances:
            return None  # No instance yet for this manual task

        # Check for incomplete fields and update status in single pass
        for inst in candidate_instances:
            if inst.incomplete_fields:
                return None  # At least one instance still incomplete

            # Update status if needed
            if inst.status != StatusType.completed:
                inst.status = StatusType.completed
                inst.save(self.resources.session)

        # Aggregate submission data from all instances
        aggregated_data = self._aggregate_submission_data(candidate_instances)

        # Merge conditional data with aggregated submission data
        if conditional_data:
            aggregated_data = deep_update(aggregated_data, conditional_data)

        return aggregated_data or None

    def _aggregate_submission_data(
        self, instances: list[ManualTaskInstance]
    ) -> dict[str, Any]:
        """Aggregate submission data from all instances into a single dictionary."""
        aggregated_data: dict[str, Any] = {}

        for inst in instances:
            # Filter valid submissions and process them
            valid_submissions = (
                submission
                for submission in inst.submissions
                if (
                    submission.field
                    and submission.field.field_key
                    and isinstance(submission.data, dict)
                )
            )

            for submission in valid_submissions:
                field_key = submission.field.field_key
                # We already checked isinstance(submission.data, dict) in valid_submissions
                data_dict: dict[str, Any] = submission.data  # type: ignore[assignment]
                field_type = data_dict.get("field_type")

                # Process field data based on type
                aggregated_data[field_key] = (
                    self._process_attachment_field(submission)
                    if field_type == ManualTaskFieldType.attachment.value
                    else data_dict.get("value")
                )

        return aggregated_data

    def _process_attachment_field(
        self, submission: ManualTaskSubmission
    ) -> Optional[list[dict[str, Any]]]:
        """Process attachment field and return attachment list or None."""
        attachment_list: list[dict[str, Any]] = []

        for attachment in filter(
            lambda a: a.attachment_type == AttachmentType.include_with_access_package,
            submission.attachments,
        ):
            try:
                size, url = attachment.retrieve_attachment()
                attachment_list.append(
                    {
                        "file_name": attachment.file_name,
                        "download_url": str(url) if url else None,
                        "file_size": size,
                    }
                )
            except Exception as exc:  # pylint: disable=broad-exception-caught
                logger.warning(
                    f"Error retrieving attachment {attachment.file_name}: {str(exc)}"
                )
        return attachment_list or None

    def _cleanup_manual_task_instances(
        self, manual_task: ManualTask, privacy_request: PrivacyRequest
    ) -> None:
        """
        Clean up ManualTaskInstances for a manual task when conditional dependencies are not met.

        This method removes any existing instances that were created before the conditional
        dependency evaluation determined the task should not execute.
        """
        # Find all instances for this manual task and privacy request
        instances_to_remove = [
            instance
            for instance in privacy_request.manual_task_instances
            if instance.task_id == manual_task.id
        ]

        if instances_to_remove:
            # Remove instances from the database
            for instance in instances_to_remove:
                instance.delete(self.resources.session)
                self.resources.session.commit()



from enum import Enum

from citext import CIText
from sqlalchemy import ARRAY, Column
from sqlalchemy import Enum as EnumColumn
from sqlalchemy import ForeignKey, String, UniqueConstraint
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship

from fides.api.db.base_class import Base
from fides.api.models.fides_user import FidesUser
from fides.api.models.sql_models import System  # type: ignore[attr-defined]


class CustomTaxonomyColor(str, Enum):
    WHITE = "taxonomy_white"
    RED = "taxonomy_red"
    ORANGE = "taxonomy_orange"
    YELLOW = "taxonomy_yellow"
    GREEN = "taxonomy_green"
    BLUE = "taxonomy_blue"
    PURPLE = "taxonomy_purple"
    SANDSTONE = "sandstone"
    MINOS = "minos"


class SystemGroup(Base):
    @declared_attr
    def __tablename__(self) -> str:
        return "system_group"

    fides_key = Column(
        String,
        ForeignKey("taxonomy_element.fides_key", ondelete="CASCADE"),
        nullable=False,
        unique=True,
        index=True,
    )
    label_color = Column(
        EnumColumn(CustomTaxonomyColor, values_callable=lambda x: [i.value for i in x]),
        nullable=True,
    )
    data_steward_username = Column(
        CIText,
        ForeignKey(FidesUser.username, ondelete="SET NULL"),
        index=True,
        nullable=True,
    )
    data_uses = Column(
        ARRAY(String),
        nullable=False,
        server_default="{}",
        default=list,
    )  # a list of `fides_key`s of `DataUse` records
    data_steward = relationship(
        "FidesUser", foreign_keys=[data_steward_username], lazy="selectin"
    )

    systems = relationship(
        System,
        secondary="system_group_member",
        back_populates="system_groups",
        viewonly=True,
    )


class SystemGroupMember(Base):
    @declared_attr
    def __tablename__(self) -> str:
        return "system_group_member"

    system_group_id = Column(
        String(255), ForeignKey("system_group.id", ondelete="CASCADE"), index=True
    )
    system_id = Column(
        String(255), ForeignKey("ctl_systems.id", ondelete="CASCADE"), index=True
    )

    __table_args__ = (
        UniqueConstraint(
            "system_group_id",
            "system_id",
            name="uq_system_group_member_group_system",
        ),
    )
